/*
 * DMA memory management for framework level HCD code (hc_driver)
 *
 * This implementation plugs in through generic "usb_bus" level methods,
 * and works with real PCI, or when "pci device == null" makes sense.
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "sys_type.h"
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_TIMER
	#include "PoTimer.h" // osport_timer.h
#endif // ndef NO_TIMER
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h"
// #endif // _ETCPIP

/*
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/pci.h>
*/
#include "PoDevice.h" // osport_device.h
#ifdef CONFIG_PCI
    #include "UsbPoPci.h" // usb_osport_pci.h
#endif	// CONFIG_PCI

#ifdef CONFIG_USB_DEBUG
	#define DEBUG
#else
	#undef DEBUG
#endif

/*
#include <linux/usb.h>
*/
#include "usb.h"

#ifndef USB_NO_DMA_POOL
	#include "UsDmaPol.h" // usb_pcipool.h
#endif // ndef USB_NO_DMA_POOL

#include "UsbHcd.h" // ../core/hcd.h


/*
 * DMA-Coherent Buffers
 */

#ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw
	/* FIXME tune these based on pool statistics ... */
	static const size_t	pool_max [HCD_BUFFER_POOLS] = {
	/* platforms without dma-friendly caches might need to
	 * prevent cacheline sharing...
	 */
	32,
	128,
	512,
	PAGE_SIZE / 2
	/* bigger --> allocate pages */
	};
#endif // ndef USB_NO_DMA_POOL


/* SETUP primitives */

/**
 * hcd_buffer_create - initialize buffer pools
 * @hcd: the bus whose buffer pools are to be initialized
 * Context: !in_interrupt()
 *
 * Call this as part of initializing a host controller that uses the pci dma
 * memory allocators.  It initializes some pools of dma-consistent memory that
 * will be shared by all drivers using that controller, or returns a negative
 * errno value on error.
 *
 * Call hcd_buffer_destroy() to clean up after using those pools.
 */
int hcd_buffer_create (struct usb_hcd *hcd)
{
 #ifdef USB_DBGSTRING_ON
	char		name [16];
 #endif // USB_DBGSTRING_ON
 int 		i, size;

 USB_DBG_FLOW(iprintf(
	"[USB FLOW] hcd_buffer_create() Begin ..\n"));
    
 #ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw
    
 for (i = 0; i < HCD_BUFFER_POOLS; i++)
   { 
	if (!(size = pool_max [i]))
       {
		continue;
       }
	USB_DBG_MEM(iprintf(
		"[USB MEM] hcd_buffer_create() : alloc hcd->pool[i] ..\n"));
        
	#ifdef USB_DBGSTRING_ON
		#ifndef USB_NO_LIB
			snprintf (name, sizeof name, "buffer-%d ", size);
		#endif // ndef USB_NO_LIB
	#endif
    
	#ifdef CONFIG_PCI
    
	#ifdef USB_DBGSTRING_ON
    	/*
		hcd->pool [i] = pci_pool_create (name, hcd->pdev,
				size, size, 0);
	    */
		hcd->pool [i] = usb_dma_pool_create (name, hcd->pdev,
				size, size,
				0 , /* no page-crossing issues */
	    	    GFP_DMA // 2003/4/18 B. S. Liaw
		        );
	  #else
        #ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw
		hcd->pool [i] = usb_dma_pool_create (hcd->pdev,
				size, size,
				0 , // no page-crossing issues 
	    	    GFP_DMA // 2003/4/18 B. S. Liaw
		        );
        #else
        hcd->pool [i] = UsbAllocMemU(USB_MEM_U_M
				#ifdef USB_DBG_MEM_U_ON
            		,size
        		#endif
                );
        #endif
	#endif
    #else
	#ifdef USB_DBGSTRING_ON
    	/*
		hcd->pool [i] = pci_pool_create (name, hcd->pdev,
				size, size, 0);
	    */
		hcd->pool [i] = usb_dma_pool_create (name,
				size, size,
				0 , /* no page-crossing issues */
	    	    GFP_DMA // 2003/4/18 B. S. Liaw
		        );
	  #else
        #ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw
		hcd->pool [i] = usb_dma_pool_create (
				size, size,
				0 , // no page-crossing issues 
	    	    GFP_DMA // 2003/4/18 B. S. Liaw
		        );
        #else
        hcd->pool [i] = 
        	UsbAllocMemU(USB_MEM_U_M,
				#ifdef USB_DBG_MEM_U_ON
            		size
		        #endif
                    );
        #endif
	#endif
    
    #endif
    
	if (!hcd->pool [i])
       {
		USB_DBG_INIT(iprintf(
			"[USB FLOW] hcd_buffer_create() : Fail , i=%d\n",i));
            
		hcd_buffer_destroy (hcd);
        
		USB_DBG_FLOW(iprintf(
			"[USB FLOW] hcd_buffer_create() End with Fail !!\n"));
		return -ENOMEM;
		}
	}
    
 #endif
    
 USB_DBG_FLOW(iprintf(
	"[USB FLOW] hcd_buffer_create() End.\n"));
 return 0;
}
#ifndef _ETCPIP
	EXPORT_SYMBOL (hcd_buffer_create);
#endif // ndef _ETCPIP


/**
 * hcd_buffer_destroy - deallocate buffer pools
 * @hcd: the bus whose buffer pools are to be destroyed
 * Context: !in_interrupt()
 *
 * This frees the buffer pools created by hcd_buffer_create().
 */
 
void hcd_buffer_destroy (struct usb_hcd *hcd)
{
 int		i;

 USB_DBG_FLOW(iprintf(
	"[USB FLOW] hcd_buffer_destroy() Begin ..\n"));
    
 #ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw

 for (i = 0; i < HCD_BUFFER_POOLS; i++)
   { 
	#ifdef CONFIG_PCI
		struct pci_pool		*pool = hcd->pool [i];
     #else
		struct dma_pool		*pool = hcd->pool [i];
    #endif
	if (pool)
       {
		#ifdef USB_NO_DMA_POOL
			UsbFreeMemU(USB_MEM_U_M, (unsigned char *)pool,
				#if defined(USB_DBG_MEM_U_ON)
                    /* 
					#ifdef CONFIG_PCI
						sizeof(struct pci_pool)
				     #else
                     
						sizeof(struct dma_pool)
                    
				    #endif
                    */
                    pool_max[i]
				#endif
		        ) ;
            #else
	            /*
				pci_pool_destroy (pool);
	            */
				usb_dma_pool_destroy (pool);
            #endif
			hcd->pool [i] = 0;
		}
	} // for i
    
 #endif   
 
 USB_DBG_FLOW(iprintf(
	"[USB FLOW] hcd_buffer_destroy() End.\n"));
}
#ifndef _ETCPIP
	EXPORT_SYMBOL (hcd_buffer_destroy);
#endif // ndef _ETCPIP


/* sometimes alloc/free could use kmalloc with SLAB_DMA, for
 * better sharing and to leverage mm/slab.c intelligence.
 */

void *hcd_buffer_alloc( struct usb_bus *bus,
	size_t		size,
	int			mem_flags,
	dma_addr_t	*dma
)
{
 struct usb_hcd		*hcd = bus->hcpriv_hcd;
 int 			i;

 USB_DBG_FLOW(iprintf(
	"[USB FLOW] hcd_buffer_alloc() Begin .. \n"
    ));
    
 #ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw
    
 for (i = 0; i < HCD_BUFFER_POOLS; i++)
   {
	if (size <= pool_max [i])
       {
		USB_DBG_INIT(iprintf(
			"[USB INIT] hcd_buffer_alloc() : alloc hcd->pool[i] .. \n"
	    	));
		#ifdef USB_NO_DMA_POOL
		   	// return usb_dma_pool_alloc(dma,size);
	        dma = UsbAllocMemU(USB_MEM_U_M,
					#ifdef USB_DBG_MEM_U_ON
            			size
                    #endif
                        );
            return dma ;
		#else
    	/*
		return pci_pool_alloc (hcd->pool [i], mem_flags, dma);
        */
		return usb_dma_pool_alloc(hcd->pool[i], mem_flags, dma);
        #endif
       }
   }
   
 #endif // ndef USB_NO_DMA_POOL
   
 USB_DBG_INIT(iprintf(
	"[USB INIT] hcd_buffer_alloc() : alloc hcd->pdev .. \n"
    ));
    
 #ifdef CONFIG_PCI
	return pci_alloc_consistent (hcd->pdev, size, dma);
  #else
	return mem_alloc_consistent(size, dma);
 #endif
}

// #ifndef _ETCPIP
   
void hcd_buffer_free (
	struct usb_bus 		*bus,
	size_t			size,
	void 			*addr,
	dma_addr_t		dma
   )
{
 struct usb_hcd		*hcd = bus->hcpriv_hcd;
 int 			i;

 USB_DBG_PLUG(iprintf("[USB PLUG] hcd_buffer_free() Begin .. \n")) ;
 #ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw
	for (i = 0; i < HCD_BUFFER_POOLS; i++) {
		if (size <= pool_max [i])
          {
			#ifdef USB_NO_DMA_POOL
		        UsbFreeMemU(USB_MEM_U_M, (unsigned char *)hcd->pool[i],
				 	#if defined(USB_DBG_MEM_U_ON)
        				pool_max [i]
					#endif
		            ) ;
            #else
				dma_pool_free (hcd->pool [i], addr, dma);
            #endif
			return;
		}
	}
 #endif // ndef USB_NO_DMA_POOL

 #ifdef CONFIG_PCI
	pci_free_consistent (hcd->pdev, size, addr, dma);
  #else
	mem_free_consistent(size, addr, dma);
 #endif
}
   
// #endif // ndef _ETCPIP
