/*
 * kobject.c - library routines for handling generic kernel objects
 *
 * Copyright (c) 2002-2003 Patrick Mochel <mochel@osdl.org>
 *
 * This file is released under the GPLv2.
 *
 *
 * Please see the file Documentation/kobject.txt for critical information
 * about using the kobject interface.
 */

#undef DEBUG

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
// #include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
// #include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP

/*
#include <linux/kobject.h>
*/
#include "kobject.h"
/*
#include <linux/string.h>
*/
#include "PoString.h"
/*
#include <linux/module.h>
#include <linux/stat.h>
*/
#include "stat.h"

// B. S. Liaw add this for pr_debug
#include "PoKernel.h"

/**
 *	populate_dir - populate directory with attributes.
 *	@kobj:	object we're working on.
 *
 *	Most subsystems have a set of default attributes that 
 *	are associated with an object that registers with them.
 *	This is a helper called during object registration that 
 *	loops through the default attributes of the subsystem 
 *	and creates attributes files for them in sysfs.
 *
 */

static int populate_dir(struct kobject * kobj)
{
 struct kobj_type * t = get_ktype(kobj);
 #ifndef NO_SYS_FS
	// struct attribute on sysfs.h
	struct attribute * attr;
 #endif // ndef NO_SYS_FS
 int error = 0;
 int i;
	
 USB_DBG_FLOW(iprintf("[USB FLOW] populate_dir() Begin .. \n"));
 #ifndef NO_SYS_FS
	// struct attribute on sysfs.h
	if (t && t->default_attrs) {
		for (i = 0; (attr = t->default_attrs[i]); i++) {
			if ((error = sysfs_create_file(kobj,attr)))
				break;
		}
	}
  #else
	error = 0 ;
 #endif
 return error;
}

#ifndef NO_SYS_FS
static int create_dir(struct kobject * kobj)
{
 int error = 0;
 if (kobject_name(kobj)) {
	#ifndef NO_SYS_FS
		// sysfs_create_dir() is on fs/sysfs/dir.c
		error = sysfs_create_dir(kobj);
		if (!error) {
			if ((error = populate_dir(kobj)))
				sysfs_remove_dir(kobj);
		   }
	#endif // ndef NO_SYS_FS
   }
 return error;
}
#endif // ndef NO_SYS_FS


static inline struct kobject * to_kobj(struct list_head * entry)
{
	return container_of(entry,struct kobject,entry);
}


#ifdef CONFIG_HOTPLUG
static int get_kobj_path_length(struct kset *kset, struct kobject *kobj)
{
	int length = 1;
	struct kobject * parent = kobj;

	/* walk up the ancestors until we hit the one pointing to the 
	 * root.
	 * Add 1 to strlen for leading '/' of each level.
	 */
	do {
		length += strlen(kobject_name(parent)) + 1;
		parent = parent->parent;
	} while (parent);
	return length;
}

static void fill_kobj_path(struct kset *kset, struct kobject *kobj, char *path, int length)
{
	struct kobject * parent;

	--length;
	for (parent = kobj; parent; parent = parent->parent) {
		int cur = strlen(kobject_name(parent));
		/* back up enough to print this name with '/' */
		length -= cur;
		strncpy (path + length, kobject_name(parent), cur);
		*(path + --length) = '/';
	}

	pr_debug("%s: path = '%s'\n",__FUNCTION__,path);
}

#define BUFFER_SIZE	1024	/* should be enough memory for the env */
#define NUM_ENVP	32	/* number of env pointers */
static unsigned long sequence_num;
static spinlock_t sequence_lock = SPIN_LOCK_UNLOCKED;

static void kset_hotplug(const char *action, struct kset *kset,
			 struct kobject *kobj)
{
	char *argv [3];
	char **envp = NULL;
	char *buffer = NULL;
	char *scratch;
	int i = 0;
	int retval;
	int kobj_path_length;
	char *kobj_path = NULL;
	char *name = NULL;
	unsigned long seq;

	/* If the kset has a filter operation, call it. If it returns
	   failure, no hotplug event is required. */
	if (kset->hotplug_ops->filter) {
		if (!kset->hotplug_ops->filter(kset, kobj))
			return;
	}

	pr_debug ("%s\n", __FUNCTION__);

	if (!hotplug_path[0])
		return;
 /*
	envp = kmalloc(NUM_ENVP * sizeof (char *), GFP_KERNEL);
 */
	envp = (char **) UsbAllocMemC(USB_MEM_C_S
					#ifdef USB_DBG_MEM_C_ON
							, NUM_ENVP * sizeof (char *)
		    	    #endif
	           		) ;
	if (!envp)
		return;
	memset (envp, 0x00, NUM_ENVP * sizeof (char *));

	/*
	buffer = kmalloc(BUFFER_SIZE, GFP_KERNEL);
	*/
	buffer = (char *) UsbAllocMemC(USB_MEM_C_S
					#ifdef USB_DBG_MEM_C_ON
							, BUFFER_SIZE
		    	    #endif
	           		) ;
	if (!buffer)
		goto exit;

	if (kset->hotplug_ops->name)
		name = kset->hotplug_ops->name(kset, kobj);
	if (name == NULL)
		name = kset->kobj.name;

	argv [0] = hotplug_path;
	argv [1] = name;
	argv [2] = 0;

	/* minimal command environment */
	envp [i++] = "HOME=/";
	envp [i++] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";

	scratch = buffer;

	envp [i++] = scratch;
	scratch += sprintf(scratch, "ACTION=%s", action) + 1;

	spin_lock(&sequence_lock);
	seq = sequence_num++;
	spin_unlock(&sequence_lock);

	envp [i++] = scratch;
	scratch += sprintf(scratch, "SEQNUM=%ld", seq) + 1;

	kobj_path_length = get_kobj_path_length (kset, kobj);
	/*
	kobj_path = kmalloc (kobj_path_length, GFP_KERNEL);
	*/
	kobj_path = (char *) UsbAllocMemC(USB_MEM_C_S
					#ifdef USB_DBG_MEM_C_ON
							, kobj_path_length
		    	    #endif
	           		) ;

	if (!kobj_path)
		goto exit;
	memset (kobj_path, 0x00, kobj_path_length);
	fill_kobj_path (kset, kobj, kobj_path, kobj_path_length);

	envp [i++] = scratch;
	scratch += sprintf (scratch, "DEVPATH=%s", kobj_path) + 1;

	if (kset->hotplug_ops->hotplug) {
		/* have the kset specific function add its stuff */
		retval = kset->hotplug_ops->hotplug (kset, kobj,
				  &envp[i], NUM_ENVP - i, scratch,
				  BUFFER_SIZE - (scratch - buffer));
		if (retval) {
			pr_debug ("%s - hotplug() returned %d\n",
				  __FUNCTION__, retval);
			goto exit;
		}
	}

	pr_debug ("%s: %s %s %s %s %s %s\n", __FUNCTION__, argv[0], argv[1],
		  envp[0], envp[1], envp[2], envp[3]);
	retval = call_usermodehelper (argv[0], argv, envp, 0);
	if (retval)
		pr_debug ("%s - call_usermodehelper returned %d\n",
			  __FUNCTION__, retval);

exit:
	kfree(kobj_path);
	kfree(buffer);
	kfree(envp);
	return;
}
#else
static void kset_hotplug(const char *action, struct kset *kset,
			 struct kobject *kobj)
{
	return;
}
#endif	/* CONFIG_HOTPLUG */

/**
 *	kobject_init - initialize object.
 *	@kobj:	object in question.
 */

void kobject_init(struct kobject * kobj)
{
 // #ifndef _ETCPIP
	atomic_set(&kobj->refcount,1);
 // #endif // ndef _ETCPIP
	INIT_LIST_HEAD(&kobj->entry);
	kobj->kset = kset_get(kobj->kset);
}


/**
 *	unlink - remove kobject from kset list.
 *	@kobj:	kobject.
 *
 *	Remove the kobject from the kset list and decrement
 *	its parent's refcount.
 *	This is separated out, so we can use it in both 
 *	kobject_del() and kobject_add() on error.
 */

static void kobj_unlink(struct kobject * kobj)
{
 #ifndef NO_KSET
	 if (kobj->kset) {
		#ifndef NO_SEMPHORE
			down_write(&kobj->kset->subsys->rwsem);
		#endif // ndef NO_SEMPHORE
		list_del_init(&kobj->entry);
		#ifndef NO_SEMPHORE
				up_write(&kobj->kset->subsys->rwsem);
		#endif // ndef NO_SEMPHORE
	}
 #endif // ndef NO_KSET
 #ifndef NO_KSET
	if (kobj->parent) 
		kobject_put(kobj->parent);
 #endif // ndef NO_KSET
 kobject_put(kobj);
}

/**
 *	kobject_add - add an object to the hierarchy.
 *	@kobj:	object.
 */

int kobject_add(struct kobject * kobj)
{
 int error = 0;
 #ifndef NO_KSET
	struct kobject * parent;
	struct kobject * top_kobj;
 #endif // ndef NO_KSET

 if (!(kobj = kobject_get(kobj)))
		return -ENOENT;

 #ifdef USB_DBGSTRING_ON
	if (!kobj->k_name)
		kobj->k_name = kobj->name;
 #endif // USB_DBGSTRING_ON
 
 #ifndef NO_KSET
	parent = kobject_get(kobj->parent);
 #endif // ndef NO_KSET

 #ifdef USB_DBGSTRING_ON
	#ifndef NO_KSET
		USB_DBG_INIT(iprintf(
			"[USB INIT] kobject_add() : kobject [%s]: registering. parent: [%s], set: [%s] \n",
			 kobject_name(kobj), parent ? kobject_name(parent) : "<NULL>", 
			 kobj->kset ? kobj->kset->kobj.name : "<NULL>" ));

		USB_DBG_INFO(iprintf(
			"[USB INFO] kobject_add() : kobject [%s]: registering. parent: [%s], set: [%s] \n",
			 kobject_name(kobj), parent ? kobject_name(parent) : "<NULL>", 
			 kobj->kset ? kobj->kset->kobj.name : "<NULL>" ));
	#else
		USB_DBG_INIT(iprintf(
			"[USB INIT] kobject_add() : kobject [%s]: registering. \n",
			 kobject_name(kobj)
			));
	#endif
  #else
		USB_DBG_INFO(iprintf(
			"[USB INFO] kobject_add() : kobject registering ..\n")) ;
 #endif

 #ifndef NO_KSET
	if (kobj->kset) {
		USB_DBG_INIT(iprintf("[USB INIT] kobject_add() : has kset .. \n"));

		#ifndef NO_SEMPHORE
			down_write(&kobj->kset->subsys->rwsem);
		#endif // ndef NO_SEMPHORE
		
		if (!parent)
				parent = kobject_get(&kobj->kset->kobj);
		
		list_add_tail(&kobj->entry,&kobj->kset->list);
		#ifndef NO_SEMPHORE
			up_write(&kobj->kset->subsys->rwsem);
		#endif // ndef NO_SEMPHORE
		}
	
	kobj->parent = parent;
	
 #endif // ndef NO_KSET

 #ifndef NO_SYS_FS
	error = create_dir(kobj);
	if (error)
		kobj_unlink(kobj);
	else {
		/* If this kobj does not belong to a kset,
		   try to find a parent that does. */
	 #ifndef NO_KSET
		top_kobj = kobj;
		if (!top_kobj->kset && top_kobj->parent) {
			do {
				top_kobj = top_kobj->parent;
			} while (!top_kobj->kset && top_kobj->parent);
		}
		if (top_kobj->kset && top_kobj->kset->hotplug_ops)
			kset_hotplug("add", top_kobj->kset, kobj);
	 #endif // ndef NO_KSET
	}
 #endif // ndef NO_SYS_FS

 USB_DBG_FLOW(iprintf("[USB FLOW] kobject_add() End. \n"));
 return error;
}


/**
 *	kobject_register - initialize and add an object.
 *	@kobj:	object in question.
 */

int kobject_register(struct kobject * kobj)
{
 int error = 0;

 if (kobj) {
	kobject_init(kobj);
	/*
	USB_DBG_INIT(iprintf("[USB INIT] kobject_register() : kobject_add() .. \n"));
	error = kobject_add(kobj);
	if (error) {
		dprintf("[USBMS] kobject_register failed for %s (%d)\n",
			       kobject_name(kobj),error);
		#ifndef _ETCPIP
			dump_stack();
		#endif // ndef _ETCPIP
		}
	*/
	} else
		error = -EINVAL;
	
 return error;
}


/**
 *	kobject_set_name - Set the name of an object
 *	@kobj:	object.
 *	@name:	name. 
 *
 *	If strlen(name) < KOBJ_NAME_LEN, then use a dynamically allocated
 *	string that @kobj->k_name points to. Otherwise, use the static 
 *	@kobj->name array.
 */

#ifdef USB_DBGSTRING_ON
int kobject_set_name(struct kobject * kobj, const char * fmt, ...)
{
 int error = 0;
 int limit = KOBJ_NAME_LEN;
 int need;
 va_list args;

 va_start(args,fmt);
 /* 
  * First, try the static array 
  */
 // #ifdef USB_DBGSTRING_ON
	need = vsnprintf(kobj->name,limit,fmt,args);
	if (need < limit) 
		kobj->k_name = kobj->name;
	else {
		/* 
		 * Need more space? Allocate it and try again 
		 */
		/*
		kobj->k_name = kmalloc(need,GFP_KERNEL);
		*/
		kobj->k_name = (char *) UsbAllocMemC(USB_MEM_C_S
									#ifdef USB_DBG_MEM_C_ON
										, need
		    					    #endif
					           		) ;

		if (!kobj->k_name) {
			error = -ENOMEM;
			goto Done;
		}
		limit = need;
		need = vsnprintf(kobj->k_name,limit,fmt,args);

		/* Still? Give up. */
		if (need > limit) {
			kfree(kobj->k_name);
			error = -EFAULT;
		}
	}
 // #endif // USB_DBGSTRING_ON

 // ---------
 Done:
 // ---------

 va_end(args);
 return error;
}
#endif // USB_DBGSTRING_ON
/*
EXPORT_SYMBOL(kobject_set_name);
*/

/**
 *	kobject_rename - change the name of an object
 *	@kobj:	object in question.
 *	@new_name: object's new name
 */

void kobject_rename(struct kobject * kobj, char *new_name)
{
 kobj = kobject_get(kobj);
 if (!kobj)
		return;
 #ifndef NO_SYS_FS
	sysfs_rename_dir(kobj, new_name);
 #endif // ndef NO_SYS_FS
 kobject_put(kobj);
}

/**
 *	kobject_del - unlink kobject from hierarchy.
 * 	@kobj:	object.
 */

void kobject_del(struct kobject * kobj)
{
 #ifndef NO_KSET
	struct kobject * top_kobj;
 #endif // ndef NO_KSET

 /* If this kobj does not belong to a kset,
	   try to find a parent that does. */
 #ifndef NO_KSET
	top_kobj = kobj;
 #endif // ndef NO_KSET

 #ifndef NO_KSET
	if (!top_kobj->kset && top_kobj->parent) {
		do {
			top_kobj = top_kobj->parent;
		} while (!top_kobj->kset && top_kobj->parent);
	}
 #endif // ndef NO_KSET
 #ifndef NO_KSET
	if (top_kobj->kset && top_kobj->kset->hotplug_ops)
		kset_hotplug("remove", top_kobj->kset, kobj);
 #endif // ndef NO_KSET

 #ifndef NO_SYS_FS
	sysfs_remove_dir(kobj);
 #endif // ndef NO_SYS_FS
 kobj_unlink(kobj);
}

/**
 *	kobject_unregister - remove object from hierarchy and decrement refcount.
 *	@kobj:	object going away.
 */

void kobject_unregister(struct kobject * kobj)
{
	pr_debug("kobject %s: unregistering\n",kobject_name(kobj));
	kobject_del(kobj);
	kobject_put(kobj);
}

/**
 *	kobject_get - increment refcount for object.
 *	@kobj:	object.
 */

struct kobject * kobject_get(struct kobject * kobj)
{
 struct kobject * ret = kobj;

 if (kobj) {
	/*
	WARN_ON(!atomic_read(&kobj->refcount));
	*/
	atomic_inc(&kobj->refcount);
   }
  else
   {
	USB_DBG_INFO(iprintf("[USB INFO] kobject_get() : Error :  Null kobj !! \n"));
	ret = NULL;
   }
 return ret;
}

/**
 *	kobject_cleanup - free kobject resources. 
 *	@kobj:	object.
 */

void kobject_cleanup(struct kobject * kobj)
{
 struct kobj_type * t = get_ktype(kobj);
 struct kset * s = kobj->kset;

 pr_debug("kobject %s: cleaning up\n",kobject_name(kobj));
 #ifdef USB_DBGSTRING_ON
	if (kobj->k_name != kobj->name)
		kfree(kobj->k_name);
	kobj->k_name = NULL;
 #endif // USB_DBGSTRING_ON
 if (t && t->release)
		t->release(kobj);
 if (s)
		kset_put(s);
}

/**
 *	kobject_put - decrement refcount for object.
 *	@kobj:	object.
 *
 *	Decrement the refcount, and if 0, call kobject_cleanup().
 */

void kobject_put(struct kobject * kobj)
{
 if (atomic_dec_and_test(&kobj->refcount))
		kobject_cleanup(kobj);
}


/**
 *	kset_init - initialize a kset for use
 *	@k:	kset 
 */

void kset_init(struct kset * k)
{
	kobject_init(&k->kobj);
	INIT_LIST_HEAD(&k->list);
}


/**
 *	kset_add - add a kset object to the hierarchy.
 *	@k:	kset.
 *
 *	Simply, this adds the kset's embedded kobject to the 
 *	hierarchy. 
 *	We also try to make sure that the kset's embedded kobject
 *	has a parent before it is added. We only care if the embedded
 *	kobject is not part of a kset itself, since kobject_add()
 *	assigns a parent in that case. 
 *	If that is the case, and the kset has a controlling subsystem,
 *	then we set the kset's parent to be said subsystem. 
 */

int kset_add(struct kset * k)
{
 if (!k->kobj.parent && !k->kobj.kset && k->subsys)
		k->kobj.parent = &k->subsys->kset.kobj;

 USB_DBG_INIT(iprintf("[USB INIT] kset_add() : kobject_add() .. \n"));
 return kobject_add(&k->kobj);
}


/**
 *	kset_register - initialize and add a kset.
 *	@k:	kset.
 */

/*
int kset_register(struct kset * k)
{
	kset_init(k);
	return kset_add(k);
}
*/

/**
 *	kset_unregister - remove a kset.
 *	@k:	kset.
 */
/*
void kset_unregister(struct kset * k)
{
	kobject_unregister(&k->kobj);
}
*/

/**
 *	kset_find_obj - search for object in kset.
 *	@kset:	kset we're looking in.
 *	@name:	object's name.
 *
 *	Lock kset via @kset->subsys, and iterate over @kset->list,
 *	looking for a matching kobject. Return object if found.
 */

struct kobject * kset_find_obj(struct kset * kset, const char * name)
{
	struct list_head * entry;
	struct kobject * ret = NULL;

 #ifndef NO_SEMPHORE
	down_read(&kset->subsys->rwsem);
 #endif // ndef NO_SEMPHORE
	list_for_each(entry,&kset->list) {
		struct kobject * k = to_kobj(entry);
		if (!strcmp(kobject_name(k),name)) {
			ret = k;
			break;
		}
	}
 #ifndef NO_SEMPHORE
	up_read(&kset->subsys->rwsem);
 #endif // ndef NO_SEMPHORE
	return ret;
}

/*
void subsystem_init(struct subsystem * s)
{
 #ifndef NO_SEMPHORE
	init_rwsem(&s->rwsem);
 #endif // ndef NO_SEMPHORE
	kset_init(&s->kset);
}
*/

/**
 *	subsystem_register - register a subsystem.
 *	@s:	the subsystem we're registering.
 *
 *	Once we register the subsystem, we want to make sure that 
 *	the kset points back to this subsystem for correct usage of 
 *	the rwsem. 
 */

/*
int subsystem_register(struct subsystem * s)
{
	int error;

 USB_DBG_FLOW(iprintf("[USB FLOW] subsystem_register() Begin .. \n"));
	subsystem_init(s);
	pr_debug("subsystem %s: registering\n",s->kset.kobj.name);

	if (!(error = kset_add(&s->kset))) {
		if (!s->kset.subsys)
			s->kset.subsys = s;
	}
	return error;
}

void subsystem_unregister(struct subsystem * s)
{
	pr_debug("subsystem %s: unregistering\n",s->kset.kobj.name);
	kset_unregister(&s->kset);
}
*/

/**
 *	subsystem_create_file - export sysfs attribute file.
 *	@s:	subsystem.
 *	@a:	subsystem attribute descriptor.
 */

int subsys_create_file(struct subsystem * s, struct subsys_attribute * a)
{
	int error = 0;
	if (subsys_get(s)) {
	#ifndef NO_SYS_FS
		// struct attribute on sysfs.h
		error = sysfs_create_file(&s->kset.kobj,&a->attr);
	 #else
		error = 0 ;
	#endif
		subsys_put(s);
	}
	return error;
}


/**
 *	subsystem_remove_file - remove sysfs attribute file.
 *	@s:	subsystem.
 *	@a:	attribute desciptor.
 */

void subsys_remove_file(struct subsystem * s, struct subsys_attribute * a)
{
 if (subsys_get(s)) {
	#ifndef NO_SYS_FS
		// struct attribute on sysfs.h
		sysfs_remove_file(&s->kset.kobj,&a->attr);
	#endif // ndef NO_SYS_FS
	subsys_put(s);
   }
}

/*
EXPORT_SYMBOL(kobject_init);
EXPORT_SYMBOL(kobject_register);
EXPORT_SYMBOL(kobject_unregister);
EXPORT_SYMBOL(kobject_get);
EXPORT_SYMBOL(kobject_put);

EXPORT_SYMBOL(kset_find_obj);

EXPORT_SYMBOL(subsystem_init);
EXPORT_SYMBOL(subsystem_register);
EXPORT_SYMBOL(subsystem_unregister);
EXPORT_SYMBOL(subsys_create_file);
EXPORT_SYMBOL(subsys_remove_file);
*/

// B. S. Liaw TEST for compiler warning !!

