/* Driver for USB Mass Storage compliant devices
 * SCSI layer glue code
 *
 * $Id: scsiglue.c,v 1.26 2002/04/22 03:39:43 mdharm Exp $
 *
 * Current development and maintenance by:
 *   (c) 1999-2002 Matthew Dharm (mdharm-usb@one-eyed-alien.net)
 *
 * Developed with the assistance of:
 *   (c) 2000 David L. Brown, Jr. (usb-storage@davidb.org)
 *   (c) 2000 Stephen J. Gowdy (SGowdy@lbl.gov)
 *
 * Initial work by:
 *   (c) 1999 Michael Gee (michael@linuxspecific.com)
 *
 * This driver is based on the 'USB Mass Storage Class' document. This
 * describes in detail the protocol used to communicate with such
 * devices.  Clearly, the designers had SCSI and ATAPI commands in
 * mind when they created this document.  The commands are all very
 * similar to commands in the SCSI-II and ATAPI specifications.
 *
 * It is important to note that in a number of cases this class
 * exhibits class-specific exemptions from the USB specification.
 * Notably the usage of NAK, STALL and ACK differs from the norm, in
 * that they are used to communicate wait, failed and OK on commands.
 *
 * Also, for certain devices, the interrupt endpoint is used to convey
 * status of a command.
 *
 * Please see http://www.one-eyed-alien.net/~mdharm/linux-usb for more
 * information about this driver.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
// #endif // _ETCPIP


/*
#include "scsiglue.h"
*/
#include "ms_scsiglue.h"
/*
#include "usb.h"
*/
#include "ms_usb.h"
/*
#include "debug.h"
*/
#include "ms_debug.h"
/*
#include "transport.h"
*/
#include "ms_transport.h"

#ifndef _ETCPIP
	#include <linux/slab.h>
	#include <linux/module.h>
#endif // ndef _ETCPIP


/***********************************************************************
 * Host functions 
 ***********************************************************************/

static const char* host_info(struct Scsi_Host *host)
{
 return "SCSI emulation for USB Mass Storage devices";
}

static int slave_configure (struct scsi_device *sdev)
{
 /* set device to use 10-byte commands where possible */
 #ifdef USBMS_SCSIGLUE_MS_6
	USBMS_DBG_INIT(iprintf("[USBMS INIT] slave_configure() : USBMS_SCSIGLUE_MS_6 .. \n")) ;
	USBMS_DBG_INIT(iprintf("[USBMS INIT] slave_configure() : Set ->use_10_for_ms=0 .. \n")) ;
	sdev->use_10_for_ms = 0;
  #else
	USBMS_DBG_INIT(iprintf("[USBMS INIT] slave_configure() : Set ->use_10_for_ms=1 .. \n")) ;
	sdev->use_10_for_ms = 1;
 #endif
 #ifdef USBMS_SCSIGLUE_RW_6
	USBMS_DBG_INIT(iprintf("[USBMS INIT] slave_configure() : USBMS_SCSIGLUE_RW_6 .. \n")) ;
	USBMS_DBG_INIT(iprintf("[USBMS INIT] slave_configure() : Set ->use_10_for_rw=0 .. \n")) ;
	sdev->use_10_for_rw = 0;
  #else
	USBMS_DBG_INIT(iprintf("[USBMS INIT] slave_configure() : Set ->use_10_for_rw = 1 .. \n")) ;
	sdev->use_10_for_rw = 1;
 #endif

 /* this is to satisify the compiler, tho I don't think the 
  * return code is ever checked anywhere. */
 return 0;
}

/* queue a command */
/* This is always called with scsi_lock(srb->host) held */
#ifdef USBMS_NO_USB_TASK	// bnchen 09.May.2005
int scsi_issue_command( Scsi_Cmnd *srb , void (*done)(Scsi_Cmnd *))
#else
static int queuecommand( Scsi_Cmnd *srb , void (*done)(Scsi_Cmnd *))
#endif
{
#ifdef USBMS_NO_USB_TASK  /* Charles 20040520 */
extern int usb_stor_task(void * __us);
#endif
 struct us_data *us = (struct us_data *)srb->device->host->hostdata[0];


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] queuecommand() Begin .. \n"));
 /*
 US_DEBUGP("%s called\n", __FUNCTION__);
 */
 srb->host_scribble = (unsigned char *)us;

 /* enqueue the command */


 #ifdef USBMS_DBG_INFO_ON
	// B. S. Liaw Add these !!
	if(us->sm_state != US_STATE_IDLE) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] queuecommand(1) : Error : Not US_STATE_IDLE : %d !! \n" ,
			us->sm_state
			)) ;
	   }
	/*
	if(us->srb != NULL) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] queuecommand(2) : Error : USB_STORAGE Error !! \n"
			)) ;
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] queuecommand(3) : Error : us->srb Not Null : %08XH !! \n",
			us->srb));
	   }
	*/
 #endif // USBMS_DBG_INFO_ON

 /*
 if(us->sm_state != US_STATE_IDLE || us->srb != NULL) {
 */
 if(us->sm_state != US_STATE_IDLE) {
	/*
	printk(KERN_ERR USB_STORAGE "Error in %s: " 
			"state = %d, us->srb = %p\n",
			__FUNCTION__, us->sm_state, us->srb);
	*/
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] queuecommand(4) : Error : USB_STORAGE Error !! \n"
		)) ;
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] queuecommand(5) : Error : state=%d, us->srb=%08XH \n",
		us->sm_state, us->srb));
	dprintf("[USBMS] STORAGE : Error : state=%d\n",
			us->sm_state);
	/*
	#ifdef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	*/
		// 2004/1/8  B. S. Liaw : BugFix for Error Recovery
		
		dprintf("[USBMS] STORAGE : Error Recovery ..\n") ;
		/*
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] queuecommand(6) : Set Value : us->srb=NULL .. \n"
			)) ;
		us->srb = NULL;
		*/
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] queuecommand(7) : Set Value : sm_state = US_STATE_IDLE .. \n"
			)) ;
		us->sm_state = US_STATE_IDLE;
		
	/*
	 #else
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] queuecommand() End with SCSI_MLQUEUE_HOST_BUSY !! \n"));
		return SCSI_MLQUEUE_HOST_BUSY;
	#endif // ndef NO_ELEVATOR
	*/
   }

 srb->scsi_done = done;
 us->srb = srb;

 USBMS_DBG_SCSI(iprintf(
	"[USBMS SCSI] queuecommand() : Set Value : us->srb=%08XH .. \n",
	us->srb
	)) ;

 /* wake up the process task */
 /*
 #ifndef NO_SEMPHORE
	up(&(us->sema));
  #else
 */
#ifdef USBMS_NO_USB_TASK  /* Charles 20040520 */
	return (usb_stor_task(us));
#else
 #ifdef USBMS_USB_WAIT_BY_EVENT     // 2004/1/2 B. S. Liaw
	/*
	relres(USBMS_RESOURCE); // add by B. S. Liaw 2003/11/29
						    // USBMS_RESOURCE is on kernel/local.h
	*/
	// 2003/11/29 B. S. Liaw : it seems to use event rather than resource
	// MT_EVT_USBMS_DOWN is on kernel/local.h
	USBMS_DBG_SCSI(iprintf("[USBMS SCSI] queuecommand() : SIGNAL(MT_EVT_USBMS_DOWN) .. \n")) ;
	USBMS_DBG_INIT(iprintf("[USBMS INIT] queuecommand() : SIGNAL(MT_EVT_USBMS_DOWN) .. \n")) ;
	// kernel/mtmacro.h : #define SIGNAL(signo) setevt(signo)
	if (SIGNAL(MT_EVT_USBMS_DOWN) != SUCCESS) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] queuecommand(8) : Error : SIGNAL MT_EVT_USBMS_DOWN=%d fail !! \n",
			MT_EVT_USBMS_DOWN
			)) ;
		dprintf("[USBMS] Error : SIGNAL MT_EVT_USBMS_DOWN %d fail !! \n",
			MT_EVT_USBMS_DOWN ) ;
	   }
  #else
	USBMS_DBG_INIT(iprintf("[USBMS INIT] queuecommand() : us=%08XH .. \n",us)) ;
	USBMS_DBG_INIT(iprintf("[USBMS INIT] queuecommand() : Set Value : us->evt_ms=1 .. \n")) ;
	// req->evt_blk_rq = 1 ;
	// BLOCK_PREEMPTION;
	us->evt_ms = 1 ;
	// UNBLOCK_PREEMPTION;
 #endif // ndef USBMS_USB_WAIT_BY_EVENT
 /*
 #endif
 */

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] queuecommand() End. \n"));
 return 0;
#endif
}

/***********************************************************************
 * Error handling functions
 ***********************************************************************/

/* Command abort */
/* This is always called with scsi_lock(srb->host) held */
static int command_abort( Scsi_Cmnd *srb )
{
 struct Scsi_Host *host = srb->device->host;
 struct us_data *us = (struct us_data *) host->hostdata[0];


 US_DEBUGP("%s called\n", __FUNCTION__);

 /* Is this command still active? */
 if(us->srb != srb) {
	US_DEBUGP ("-- nothing to abort\n");
	return FAILED;
   }

 /* Normally the current state is RUNNING.  If the control thread
  * hasn't even started processing this command, the state will be
  * IDLE.  Anything else is a bug. */
 if(   us->sm_state != US_STATE_RUNNING
	&& us->sm_state != US_STATE_IDLE) {
	/*
	printk(KERN_ERR USB_STORAGE "Error in %s: "
			"invalid state %d\n", __FUNCTION__, us->sm_state);
	*/
	/*
	dprintf("[USBMS] Error : USB_STORAGE Error in %s: "
			"invalid state %d\n", __FUNCTION__, us->sm_state);
	*/
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] command_abort() : Error : invalid state %d !! \n" ,
		us->sm_state
		)) ;
	dprintf("[USBMS] USB_STORAGE : Error : invalid state %d\n",
			us->sm_state);
	return FAILED;
   }

 /* Set state to ABORTING, set the ABORTING bit, and release the lock */
 us->sm_state = US_STATE_ABORTING;
 USBMS_DBG_INFO(iprintf(
	"[USBMS INFO] command_abort() : Error : US_STATE_ABORTING !! \n"
	)) ;
 set_bit(US_FLIDX_ABORTING, &us->flags);
 /*
 #ifndef NO_SPINLOCK
	scsi_unlock(host);
 #endif // ndef NO_SPINLOCK
 */
 /* Stop an ongoing USB transfer */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] command_abort() : usb_stor_stop_transport() ..\n" )) ;
 usb_stor_stop_transport(us);

 /* Wait for the aborted command to finish */
 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	wait_for_completion(&us->notify);
 #endif // ndef _ETCPIP
 */
 /* Reacquire the lock and allow USB transfers to resume */
 /*
 #ifndef NO_SPINLOCK
	scsi_lock(host);
 #endif // ndef NO_SPINLOCK
 */
 clear_bit(US_FLIDX_ABORTING, &us->flags);
 return SCSI_SUCCESS;
}

/* This invokes the transport reset mechanism to reset the state of the
 * device */
/* This is always called with scsi_lock(srb->host) held */
static int device_reset( Scsi_Cmnd *srb )
{
 struct us_data *us = (struct us_data *)srb->device->host->hostdata[0];
 int result;


 US_DEBUGP("%s called\n", __FUNCTION__);
 if(us->sm_state != US_STATE_IDLE) {
	/*
	printk(KERN_ERR USB_STORAGE "Error in %s: "
		"invalid state %d\n", __FUNCTION__, us->sm_state);
	*/
	/*
	dprintf("[USBMS] Error :USB_STORAGE Error in %s: "
			"invalid state %d\n", __FUNCTION__, us->sm_state);
	*/
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] device_reset() : Error : Not US_STATE_IDLE : %d !! \n" ,
		us->sm_state
		)) ;
	return FAILED;
   }

 /* set the state and release the lock */
 us->sm_state = US_STATE_RESETTING;

 USBMS_DBG_INFO(iprintf(
	"[USBMS INFO] device_reset() : Error : US_STATE_RESETTING !! \n"
	)) ;

 /*
 #ifndef NO_SPINLOCK
	scsi_unlock(srb->device->host);
 #endif // ndef NO_SPINLOCK
 */
 /* lock the device pointers and do the reset */
 /*
 #ifndef NO_SEMPHORE
	down(&(us->dev_semaphore));
 #endif // ndef NO_SEMPHORE
 */
 if(test_bit(US_FLIDX_DISCONNECTING, &us->flags)) {
	result = FAILED;
	// US_DEBUGP("No reset during disconnect\n");
	USBMS_DBG_PLUG(iprintf(
		"[USBMS PLUG] device_reset() : No reset during disconnect !! \n"
		)) ;
   } else {
	dprintf("[USBMS] Error : SCSI Error , Reset .. \n") ;
	result = us->transport_reset(us);
   }
 /* unlock */
 /*
 #ifndef NO_SEMPHORE
	up(&(us->dev_semaphore));
 #endif // ndef NO_SEMPHORE
 */
 /* lock access to the state and clear it */
 /*
 #ifndef NO_SPINLOCK
	scsi_lock(srb->device->host);
 #endif // ndef NO_SPINLOCK
 */
 us->sm_state = US_STATE_IDLE;
 return result;
}

/* This resets the device's USB port. */
/* It refuses to work if there's more than one interface in
 * the device, so that other users are not affected. */
/* This is always called with scsi_lock(srb->host) held */
static int bus_reset( Scsi_Cmnd *srb )
{
 struct us_data *us = (struct us_data *)srb->device->host->hostdata[0];
 int result;


 US_DEBUGP("%s called\n", __FUNCTION__);
 if (us->sm_state != US_STATE_IDLE) {
	/*
	printk(KERN_ERR USB_STORAGE "Error in %s: "
			"invalid state %d\n", __FUNCTION__, us->sm_state);
	*/
	/*
	dprintf("[USBMS] Error :USB_STORAGE Error in %s: "
			"invalid state %d\n", __FUNCTION__, us->sm_state);
	*/
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] bus_reset() : Error : Not US_STATE_IDLE : %d !! \n" ,
		us->sm_state
		)) ;
	return FAILED;
   }

 /* set the state and release the lock */
 us->sm_state = US_STATE_RESETTING;

 USBMS_DBG_INFO(iprintf(
	"[USBMS INFO] bus_reset() : Error : US_STATE_RESETTING !! \n"
	)) ;

 /*
 #ifndef NO_SPINLOCK
	scsi_unlock(srb->device->host);
 #endif // ndef NO_SPINLOCK
 */
 /* The USB subsystem doesn't handle synchronisation between
  * a device's several drivers. Therefore we reset only devices
  * with just one interface, which we of course own. */
	
 #ifndef NO_SEMPHORE
	down(&(us->dev_semaphore));
 #endif // ndef NO_SEMPHORE
	if (test_bit(US_FLIDX_DISCONNECTING, &us->flags)) {
		result = -EIO;
		/*
		US_DEBUGP("No reset during disconnect\n");
		*/
		USBMS_DBG_PLUG(iprintf(
			"[USBMS PLUG] bus_reset() : No rese during disconnect !! \n"
			)) ;
	} else if (us->pusb_dev->actconfig->desc.bNumInterfaces != 1) {
		result = -EBUSY;
		US_DEBUGP("Refusing to reset a multi-interface device\n");
	} else {
		result = usb_reset_device(us->pusb_dev);
		US_DEBUGP("usb_reset_device returns %d\n", result);
	}
 #ifndef NO_SEMPHORE
	up(&(us->dev_semaphore));
 #endif // ndef NO_SEMPHORE

 /* lock access to the state and clear it */
 /*
 #ifndef NO_SPINLOCK
	scsi_lock(srb->device->host);
 #endif // ndef NO_SPINLOCK
 */
 us->sm_state = US_STATE_IDLE;
 return result < 0 ? FAILED : SCSI_SUCCESS;
}

/***********************************************************************
 * /proc/scsi/ functions
 ***********************************************************************/

/* we use this macro to help us write into the buffer */
#undef SPRINTF
#define SPRINTF(args...) \
	do { if (pos < buffer+length) pos += sprintf(pos, ## args); } while (0)
/*
#ifndef _ETCPIP
static int proc_info (struct Scsi_Host *hostptr, char *buffer, char **start, off_t offset,
		int length, int inout)
{
	struct us_data *us;
	char *pos = buffer;
	unsigned long f;

	// if someone is sending us data, just throw it away 
	if (inout)
		return length;

	us = (struct us_data*)hostptr->hostdata[0];

	// print the controller name 
	SPRINTF("   Host scsi%d: usb-storage\n", hostptr->host_no);

	// print product, vendor, and serial number strings 
	SPRINTF("       Vendor: %s\n", us->vendor);
	SPRINTF("      Product: %s\n", us->product);
	SPRINTF("Serial Number: %s\n", us->serial);

	// show the protocol and transport 
	SPRINTF("     Protocol: %s\n", us->protocol_name);
	SPRINTF("    Transport: %s\n", us->transport_name);

	// show the device flags 
	if (pos < buffer + length) {
		pos += sprintf(pos, "       Quirks:");
		f = us->flags;

#define DO_FLAG(a)  	if (f & US_FL_##a)  pos += sprintf(pos, " " #a)
		DO_FLAG(SINGLE_LUN);
		DO_FLAG(SCM_MULT_TARG);
		DO_FLAG(FIX_INQUIRY);
		DO_FLAG(FIX_CAPACITY);
#undef DO_FLAG

		*(pos++) = '\n';
		}

	
	 // Calculate start of next buffer, and return value.
	  
	*start = buffer + offset;

	if ((pos - buffer) < offset)
		return (0);
	else if ((pos - buffer - offset) < length)
		return (pos - buffer - offset);
	else
		return (length);
}
#endif // ndef _ETCPIP
*/
/*
 * this defines our host template, with which we'll allocate hosts
 */

struct scsi_host_template usb_stor_host_template = {
 /* basic userland interface stuff */
 #ifdef USB_DBGSTRING_ON
	.name =				"usb-storage",
	.proc_name =		"usb-storage",
 #endif // USB_DBGSTRING_ON 
 /*
 #ifndef _ETCPIP
 .proc_info =			proc_info,
 #endif // ndef _ETCPIP
 */
 .info =				host_info,

 /* command interface -- queued only */
#ifdef USBMS_NO_USB_TASK	// bnchen 09.May.2005
 .queuecommand =			scsi_issue_command,
#else
 .queuecommand =			queuecommand,
#endif

 /* error and abort handlers */
 .eh_abort_handler =		command_abort,
 .eh_device_reset_handler =	device_reset,
 .eh_bus_reset_handler =		bus_reset,

 /* queue commands only, only one command per LUN */
 .can_queue =			1,
 .cmd_per_lun =			1,

 /* unknown initiator id */
 .this_id =			-1,

 .slave_configure =		slave_configure,

 /* lots of sg segments can be handled */
 .sg_tablesize =			SG_ALL,

 /* merge commands... this seems to help performance, but
  * periodically someone should test to see which setting is more
  * optimal.
  */
 .use_clustering =		TRUE,

	/* emulated HBA */
	.emulated =			TRUE,

	/* module management */
 /*
 #ifndef _ETCPIP
	.module =			THIS_MODULE
 #endif // ndef _ETCPIP
 */
};

/* For a device that is "Not Ready" */
unsigned char usb_stor_sense_notready[18] = {
	[0]	= 0x70,			    /* current error */
	[2]	= 0x02,			    /* not ready */
	[7]	= 0x0a,			    /* additional length */
	[12]	= 0x04,			    /* not ready */
	[13]	= 0x03			    /* manual intervention */
};

/* To Report "Illegal Request: Invalid Field in CDB" */
unsigned char usb_stor_sense_invalidCDB[18] = {
	[0]	= 0x70,			    /* current error */
	[2]	= ILLEGAL_REQUEST,	    /* Illegal Request = 0x05 */
	[7]	= 0x0a,			    /* additional length */
	[12]	= 0x24			    /* Invalid Field in CDB */
};

