/*
 * scsi_scan.c
 *
 * Copyright (C) 2000 Eric Youngdale,
 * Copyright (C) 2002 Patrick Mansfield
 *
 * The general scanning/probing algorithm is as follows, exceptions are
 * made to it depending on device specific flags, compilation options, and
 * global variable (boot or module load time) settings.
 *
 * A specific LUN is scanned via an INQUIRY command; if the LUN has a
 * device attached, a Scsi_Device is allocated and setup for it.
 *
 * For every id of every channel on the given host:
 *
 * 	Scan LUN 0; if the target responds to LUN 0 (even if there is no
 * 	device or storage attached to LUN 0):
 *
 * 		If LUN 0 has a device attached, allocate and setup a
 * 		Scsi_Device for it.
 *
 * 		If target is SCSI-3 or up, issue a REPORT LUN, and scan
 * 		all of the LUNs returned by the REPORT LUN; else,
 * 		sequentially scan LUNs up until some maximum is reached,
 * 		or a LUN is seen that cannot have a device attached to it.
 */

// #ifdef _ETCPIP
#define __KERNEL__
        #ifdef _NEWLIB
          #include "string.h"
        #endif
   	#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
       // for deciding whether to compile
       //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"


#include "PoList.h" // osport_list.h

#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
#include "UsMemLib.h" // UsbMemLib.h
// #endif // _ETCPIP


#ifndef _ETCPIP
	#include <linux/config.h>
	#include <linux/module.h>
    #include <linux/moduleparam.h>
	#include <linux/init.h>
#endif // ndef _ETCPIP
/*
#include <linux/blkdev.h>
*/
#include "bitops.h" // B. S. Liaw add for test_bit()
#include "blkdev.h"
/*
#include <asm/semaphore.h>
*/

#include "scsi_ds.h"
#include "scsi_hosts.h"
/*
#include <scsi/scsi_driver.h>
*/
#include "scsi_driver.h"

#include "scsi_priv.h"
#include "scsi_logging.h"
#include "scsi_devinfo.h"

#include "config.h"
#include "nbios.h"

/*
#define ALLOC_FAILURE_MSG	KERN_ERR "%s: Allocation failure during" \
	" SCSI scanning, some SCSI devices might not be configured\n"
*/
/*
#define ALLOC_FAILURE_MSG "[USBMS] Error : %s: Allocation failure during" \
	" SCSI scanning, some SCSI devices might not be configured\n"
*/
#define ALLOC_FAILURE_MSG "[USBMS] Error : Out of memory during" \
	" SCSI scanning, some SCSI devices might not be configured\n"

/*
 * Prefix values for the SCSI id's (stored in driverfs name field)
 */
#define SCSI_UID_SER_NUM 'S'
#define SCSI_UID_UNKNOWN 'Z'

/*
 * Return values of some of the scanning functions.
 *
 * SCSI_SCAN_NO_RESPONSE: no valid response received from the target, this
 * includes allocation or general failures preventing IO from being sent.
 *
 * SCSI_SCAN_TARGET_PRESENT: target responded, but no device is available
 * on the given LUN.
 *
 * SCSI_SCAN_LUN_PRESENT: target responded, and a device is available on a
 * given LUN.
 */
#define SCSI_SCAN_NO_RESPONSE		0
#define SCSI_SCAN_TARGET_PRESENT	1
#define SCSI_SCAN_LUN_PRESENT		2

static char *scsi_null_device_strs = "nullnullnullnull";

#define MAX_SCSI_LUNS	512

#ifdef CONFIG_SCSI_MULTI_LUN
	static unsigned int max_scsi_luns = MAX_SCSI_LUNS;
#else
	static unsigned int max_scsi_luns = 1 ;
#endif

/*
#ifndef _ETCPIP
module_param_named(max_luns, max_scsi_luns, int, S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(max_luns,
		 "last scsi LUN (should be between 1 and 2^32-1)");
#endif // ndef _ETCPIP
*/

#ifdef CONFIG_SCSI_REPORT_LUNS
/*
 * max_scsi_report_luns: the maximum number of LUNS that will be
 * returned from the REPORT LUNS command. 8 times this value must
 * be allocated. In theory this could be up to an 8 byte value, but
 * in practice, the maximum number of LUNs suppored by any device
 * is about 16k.
 */
/*
static unsigned int max_scsi_report_luns = 128;
*/
static unsigned int max_scsi_report_luns = MAX_REPORT_LUNS ;

/*
#ifndef _ETCPIP
module_param_named(max_report_luns, max_scsi_report_luns, int, S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(max_report_luns,
		 "REPORT LUNS maximum number of LUNS received (should be"
		 " between 1 and 16384)");
#endif // ndef _ETCPIP
*/
#endif

/**
 * scsi_unlock_floptical - unlock device via a special MODE SENSE command
 * @sreq:	used to send the command
 * @result:	area to store the result of the MODE SENSE
 *
 * Description:
 *     Send a vendor specific MODE SENSE (not a MODE SELECT) command using
 *     @sreq to unlock a device, storing the (unused) results into result.
 *     Called for BLIST_KEY devices.
 **/
static void scsi_unlock_floptical(struct scsi_request *sreq,
				  unsigned char *result)
{
 unsigned char scsi_cmd[MAX_COMMAND_SIZE];


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_unlock_floptical() Begin .. \n"));
	/*
	printk(KERN_NOTICE "scsi: unlocking floptical drive\n");
	*/
	dprintf("[USBMS] NOTICE scsi: unlocking floptical drive\n");

	scsi_cmd[0] = MODE_SENSE;
	scsi_cmd[1] = 0;
	scsi_cmd[2] = 0x2e;
	scsi_cmd[3] = 0;
	scsi_cmd[4] = 0x2a;	/* size */
	scsi_cmd[5] = 0;
	sreq->sr_cmd_len = 0;
	sreq->sr_data_direction = DMA_FROM_DEVICE;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_unlock_floptical() : scsi_wait_req() .. \n"));
 scsi_wait_req(sreq, scsi_cmd, result, 0x2a /* size */, SCSI_TIMEOUT, 3);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_unlock_floptical() End. \n"));
}

/**
 * print_inquiry - printk the inquiry information
 * @inq_result:	printk this SCSI INQUIRY
 *
 * Description:
 *     printk the vendor, model, and other information found in the
 *     INQUIRY data in @inq_result.
 *
 * Notes:
 *     Remove this, and replace with a hotplug event that logs any
 *     relevant information.
 **/
static void print_inquiry(unsigned char *inq_result)
{
	int i;
	char vendorTmp[12], modelTmp[20];


	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] print_inquiry() Begin .. \n")) ;
	memset(vendorTmp, 0, 12);
	memset(modelTmp, 0, 20);

	//printk(KERN_NOTICE "  Vendor: ");
	dprintf("[USBMS] Vendor: ");
	for (i = 8; i < 16; i++) {
#ifdef USBMS_SCSI_LOOSER_CHECK
		if (inq_result[i] >= 0x20 ) {
#else
		if (inq_result[i] >= 0x20 && i < inq_result[4] + 5) {
#endif
			printk("%c", inq_result[i]);
			vendorTmp[i - 8] = inq_result[i];
		}
		else {
			printk(" ");
			vendorTmp[i - 8] = ' ';
		}
	}

	//printk("  Model: ");
	dprintf("  Model: ");
	for (i = 16; i < 32; i++) {
#ifdef USBMS_SCSI_LOOSER_CHECK
		if (inq_result[i] >= 0x20 ) {
#else
		if (inq_result[i] >= 0x20 && i < inq_result[4] + 5) {
#endif
			printk("%c", inq_result[i]);
			modelTmp[i - 16] = inq_result[i];
		}
		else {
			printk(" ");
			modelTmp[i - 16] = ' ';
		}
	}

	//printk("  Rev: ");
	dprintf("  Rev: ");
	for(i = 32; i < 36; i++) {
#ifdef USBMS_SCSI_LOOSER_CHECK
		if (inq_result[i] >= 0x20 )
#else
		if (inq_result[i] >= 0x20 && i < inq_result[4] + 5)
#endif
			printk("%c", inq_result[i]);
		else
			printk(" ");
	}

	printk("\n");

	i = inq_result[0] & 0x1f;
	//printk(KERN_NOTICE "  Type:   %s ",
	//		i <
	//		MAX_SCSI_DEVICE_CODE ? scsi_device_types[i] :
	//		"Unknown          ");
	dprintf("[USBMS] Type:   %s ", ((i < MAX_SCSI_DEVICE_CODE) ? scsi_device_types[i] : "Unknown          "));
	printk("                 ANSI SCSI revision: %02x", inq_result[2] & 0x07);
	if ((inq_result[2] & 0x07) == 1 && (inq_result[3] & 0x0f) == 1)
		printk(" CCS\n");
	else
		printk("\n");

	// update drive information for resource share
#if defined(_NB_SHARE) && defined(_NB_FILE)
	updateDiskDriveInfo(vendorTmp, modelTmp, DEFAULT_DISK_DRIVE_ID);
#endif // defined(_NB_SHARE) && defined(_NB_FILE)

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] print_inquiry() End. \n")) ;
}

/**
 * scsi_alloc_sdev - allocate and setup a scsi_Device
 *
 * Description:
 *     Allocate, initialize for io, and return a pointer to a scsi_Device.
 *     Stores the @shost, @channel, @id, and @lun in the scsi_Device, and
 *     adds scsi_Device to the appropriate list.
 *
 * Return value:
 *     scsi_Device pointer, or NULL on failure.
 **/
static struct scsi_device *scsi_alloc_sdev(struct Scsi_Host *shost,
	       	uint channel, uint id, uint lun)
{
 struct scsi_device *sdev, *device;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_sdev() Begin .. \n"));
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_alloc_sdev() : sdev = UsbAllocMemC() , size=%d .. \n" ,
	sizeof(struct scsi_device)
	));

 /*
 sdev = kmalloc(sizeof(*sdev), GFP_ATOMIC);
 */
 sdev = (struct scsi_device *) UsbAllocMemC(USB_MEM_C_M
							#ifdef USB_DBG_MEM_C_ON
								, sizeof(struct scsi_device) // size : 432
						    #endif
					    	) ;

 if (!sdev)
   {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_alloc_sdev() : Error : Out of Memory for sdev !! \n")) ;
	dprintf("[USBMS] Error : Out of Memory !! \n") ;
	goto out;
   }

 /*
 memset(sdev, 0, sizeof(*sdev));
 */
 atomic_set(&sdev->access_count, 0);
 sdev->vendor = scsi_null_device_strs;
 sdev->model = scsi_null_device_strs;
 sdev->rev = scsi_null_device_strs;
 sdev->host = shost;
 sdev->id = id;
 USBMS_DBG_SCAN(iprintf(
	"[USBMS SCAN] scsi_alloc_sdev() : Set Value : sdev->id=%d \n" ,
	sdev->id
	));
 sdev->lun = lun;
 sdev->channel = channel;
 sdev->online = TRUE;
 INIT_LIST_HEAD(&sdev->siblings);
 INIT_LIST_HEAD(&sdev->same_target_siblings);
 #ifndef NO_ELEVATOR // struct elevator_s , // include/linux/elevator.h 
	INIT_LIST_HEAD(&sdev->cmd_list);
 #endif // ndef NO_ELEVATOR
 INIT_LIST_HEAD(&sdev->starved_entry);

 /*
 #ifndef NO_SPINLOCK
	spin_lock_init(&sdev->list_lock);
 #endif // ndef NO_SPINLOCK
 */
 /*
  * Some low level driver could use device->type
  */
 sdev->type = -1;

 /*
  * Assume that the device will have handshaking problems,
  * and then fix this field later if it turns out it
  * doesn't
  */
 sdev->borken = 1;
 /*
 #ifndef NO_SPINLOCK
	spin_lock_init(&sdev->sdev_lock);
 #endif // ndef NO_SPINLOCK
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_sdev() : scsi_alloc_queue() .. \n"));
 sdev->request_queue = scsi_alloc_queue(sdev);
 if (!sdev->request_queue)
		goto out_free_dev;

 sdev->request_queue->queuedata = sdev;
 scsi_adjust_queue_depth(sdev, 0, sdev->host->cmd_per_lun);

 if (shost->hostt->slave_alloc) {
	if (shost->hostt->slave_alloc(sdev))
			goto out_free_queue;
	}

 /*
  * If there are any same target siblings, add this to the
  * sibling list
  */
 #ifndef _ETCPIP
	list_for_each_entry(device, &shost->my_devices, siblings) {
  #else
	list_for_each_entry(device, shost->my_devices, siblings) {
 #endif
	if(device->id == sdev->id &&
	    device->channel == sdev->channel) {
		USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_alloc_sdev() : device->same_target_siblings=%08XH \n",
			device->same_target_siblings ));
		#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) &device->same_target_siblings
				   > (unsigned long) 0x200000L
		         )
			   )
			   {
				iprintf("[ASSERT][USBMS] scsi_alloc_sdev() : " ) ;
    		    iprintf("Error : Invlid address &device->same_target_siblings=%08XH \n",
						&device->same_target_siblings) ;
			    iprintf("[ASSERT][USBMS] Continue ..\n") ;
    		   }
		#endif // USBMS_BS_ASSERT
			list_add_tail(&sdev->same_target_siblings,
				      &device->same_target_siblings);
			sdev->scsi_level = device->scsi_level;
			break;
	   }
   }

 /*
  * If there wasn't another lun already configured at this
  * target, then default this device to SCSI_2 until we
  * know better
  */
 if (!sdev->scsi_level)
		sdev->scsi_level = SCSI_2;

 /*
  * Add it to the end of the shost->my_devices list.
  */
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_alloc_sdev() : shost->my_devices=%08XH \n",
			shost->my_devices ));
		#ifdef USBMS_BS_ASSERT
			if (!(   (unsigned long) &shost->my_devices
				   > (unsigned long) 0x200000L
		         )
			   )
			   {
				iprintf("[ASSERT][USBMS] scsi_alloc_sdev() : " ) ;
    		    iprintf("Error : Invlid address &shost->my_devices=%08XH \n",
						&shost->my_devices) ;
			    iprintf("[ASSERT][USBMS] Continue ..\n") ;
    		   }
		#endif // USBMS_BS_ASSERT
 /*
 #ifndef _ETCPIP
	list_add_tail(&sdev->siblings, &shost->my_devices);
  #else
 */
	list_add_tail(&sdev->siblings, shost->my_devices);
 /*
 #endif
 */

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_sdev() End.\n"));
 return sdev;

 out_free_queue:
		scsi_free_queue(sdev->request_queue);
 out_free_dev:

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_alloc_sdev() : UsbFreeMemC(sdev) .. \n"
		));
 /*
 kfree(sdev);
 */
 UsbFreeMemC(USB_MEM_C_M,(unsigned char *) sdev
		#if defined(USB_DBG_MEM_C_ON)
				,sizeof(struct scsi_device) // size : 432
		#endif
		);
 out:
 /*
 printk(ALLOC_FAILURE_MSG, __FUNCTION__);
 */
 dprintf(ALLOC_FAILURE_MSG);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_alloc_sdev() End with Fail !! \n"));
 return NULL;
}

/**
 * scsi_free_sdev - cleanup and free a scsi_device
 * @sdev:	cleanup and free this scsi_device
 *
 * Description:
 *     Undo the actions in scsi_alloc_sdev, including removing @sdev from
 *     the list, and freeing @sdev.
 **/
void scsi_free_sdev(struct scsi_device *sdev)
{
 unsigned long flags;


 // USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_free_sdev() Begin .. \n"));
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_free_sdev() Begin .. \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_free_sdev() : list_del() .. \n"));
 list_del(&sdev->siblings);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_free_sdev() : list_del() .. \n"));
 list_del(&sdev->same_target_siblings);

 if (sdev->request_queue)
		scsi_free_queue(sdev->request_queue);
	if (sdev->inquiry)
       {
		USBMS_DBG_MEM(iprintf(
			"[USBMS MEM] scsi_free_sdev() : UsbFreeMemC(sdev->inquiry) .. \n"
			));
		//	kfree(sdev->inquiry);
		UsbFreeMemC(USB_MEM_C_S,(unsigned char *) sdev->inquiry
					#if defined(USB_DBG_MEM_C_ON)
						,sizeof(*sdev->inquiry)
					#endif
					);
       }
 /*
 #ifndef NO_SPINLOCK
	spin_lock_irqsave(sdev->host->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_free_sdev() : list_del() .. \n"));
 list_del(&sdev->starved_entry);
 if (sdev->single_lun) {
		if (--sdev->sdev_target->starget_refcnt == 0)
			kfree(sdev->sdev_target);
	}
 /*
 #ifndef NO_SPINLOCK
	spin_unlock_irqrestore(sdev->host->host_lock, flags);
 #endif // ndef NO_SPINLOCK
 */

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_free_sdev() : UsbFreeMemC(sdev) .. \n"
		));
 /*
 kfree(sdev);
 */
 UsbFreeMemC(USB_MEM_C_M,(unsigned char *) sdev
		#if defined(USB_DBG_MEM_C_ON)
				,sizeof(struct scsi_device) // size : 432
		#endif
		);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_free_sdev() End. \n"));
}

/**
 * scsi_probe_lun - probe a single LUN using a SCSI INQUIRY
 * @sreq:	used to send the INQUIRY
 * @inq_result:	area to store the INQUIRY result
 * @bflags:	store any bflags found here
 *
 * Description:
 *     Probe the lun associated with @sreq using a standard SCSI INQUIRY;
 *
 *     If the INQUIRY is successful, sreq->sr_result is zero and: the
 *     INQUIRY data is in @inq_result; the scsi_level and INQUIRY length
 *     are copied to the Scsi_Device at @sreq->sr_device (sdev);
 *     any flags value is stored in *@bflags.
 **/
static void scsi_probe_lun(struct scsi_request *sreq, char *inq_result,
			   int *bflags)
{
 struct scsi_device *sdev = sreq->sr_device;	/* a bit ugly */
 unsigned char scsi_cmd[MAX_COMMAND_SIZE];
 int possible_inq_resp_len;
#ifdef USBMS_DBG_INFO_ON
 int i;
#endif

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_lun() Begin .. \n"));
 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_probe_lun() : sdev=%08XH \n",
	sdev));

 *bflags = 0;

 // ----------------
 repeat_inquiry:
 // ----------------

 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_probe_lun() : sdev=%08X \n" ,
	sdev )) ;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_probe_lun() : sdev->host=%08X \n" ,
	sdev->host )) ;
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_probe_lun() : sdev->host->host_no=%08X \n" ,
	sdev->host->host_no )) ;
 /*
 SCSI_LOG_SCAN_BUS(3, printk(KERN_INFO "scsi scan: INQUIRY to host %d"
 */
 USBMS_DBG_SCAN(iprintf(
	"[USBMS SCAN] scsi_probe_lun() : INQUIRY to host %d"
	" channel %d id %d lun %d\n", sdev->host->host_no,
	sdev->channel, sdev->id, sdev->lun));

 memset(scsi_cmd, 0, 6);
 scsi_cmd[0] = INQUIRY;
 scsi_cmd[4] = 36;	/* issue conservative alloc_length */
 sreq->sr_cmd_len = 0;
 sreq->sr_data_direction = DMA_FROM_DEVICE;

 memset(inq_result, 0, 36);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_lun(1) : scsi_wait_req() .. \n"));
 scsi_wait_req(sreq, (void *) scsi_cmd, (void *) inq_result, 36,
				/*
		      	SCSI_TIMEOUT + 4 * HZ, 3);
				*/
		      SCSI_TIMEOUT + 4 , 3);
 /*
 SCSI_LOG_SCAN_BUS(3, printk(KERN_INFO "scsi scan: 1st INQUIRY %s with"
 */
 USBMS_DBG_SCAN(iprintf("[USBMS SCAN] scsi_probe_lun() : 1st INQUIRY %s with"
			" code 0x%x\n", sreq->sr_result ?
			"failed" : "successful", sreq->sr_result));

 if (sreq->sr_result) {
	if ((driver_byte(sreq->sr_result) & DRIVER_SENSE) != 0 &&
		    (sreq->sr_sense_buffer[2] & 0xf) == UNIT_ATTENTION &&
		    sreq->sr_sense_buffer[12] == 0x28 &&
		    sreq->sr_sense_buffer[13] == 0) {
		/* not-ready to ready transition - good */
		/* dpg: bogus? INQUIRY never returns UNIT_ATTENTION */
	   } else
	   {
		/*
		 * assume no peripheral if any other sort of error
		 */
		USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_probe_lun() : assume no peripheral !! \n"));
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_lun(2) End with Fail !! \n"));
		return;
	   }
   }

 /*
  * Get any flags for this device.
  *
  * XXX add a bflags to Scsi_Device, and replace the corresponding
  * bit fields in Scsi_Device, so bflags need not be passed as an
  * argument.
  */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_lun(3) : scsi_get_device_flags() .. \n"));
 
#ifdef USBMS_DBG_INFO_ON
 for (i=0; i < 36; i++)
 {
 	iprintf("%02X ", inq_result[i]);
 	if ((i & 0x0f) == 0x0f)
 		iprintf("\n");
 }
 iprintf("\n");
#endif

 *bflags |= scsi_get_device_flags(&inq_result[8], &inq_result[16]);

 possible_inq_resp_len = (unsigned char) inq_result[4] + 5;
 if (BLIST_INQUIRY_36 & *bflags)
		possible_inq_resp_len = 36;
 else if (BLIST_INQUIRY_58 & *bflags)
		possible_inq_resp_len = 58;
 else if (possible_inq_resp_len > 255)
		possible_inq_resp_len = 36;	/* sanity */
 else if (possible_inq_resp_len <= 5)
   {
	// B. S. Liaw add
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] scsi_probe_lun() : Set BLIST_INQUIRY_36 .. \n"
		));
	*bflags = BLIST_INQUIRY_36;
	possible_inq_resp_len = 36;
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] scsi_probe_lun() : Set Value : possible_inq_resp_len=%d .. \n",
		possible_inq_resp_len
		));
   }
 else if (possible_inq_resp_len == 37)  /* For ViPowEr SmartHandy Disk */
 	possible_inq_resp_len = 36;

 USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] scsi_probe_lun() : possible_inq_resp_len=%d .. \n",
		possible_inq_resp_len
		));

 if(possible_inq_resp_len > 36) {	/* do additional INQUIRY */
	memset(scsi_cmd, 0, 6);
	scsi_cmd[0] = INQUIRY;
	scsi_cmd[4] = (unsigned char) possible_inq_resp_len;
	sreq->sr_cmd_len = 0;
	sreq->sr_data_direction = DMA_FROM_DEVICE;
	/*
	 * re-zero inq_result just to be safe.
	 */
	memset(inq_result, 0, possible_inq_resp_len);

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_lun(4) : scsi_wait_req() .. \n"));
	scsi_wait_req(sreq, (void *) scsi_cmd,
			      (void *) inq_result,
				/*
			      possible_inq_resp_len, SCSI_TIMEOUT + 4 * HZ, 3);
				*/
			      possible_inq_resp_len, SCSI_TIMEOUT + 4, 3);
	/*
	SCSI_LOG_SCAN_BUS(3, printk(KERN_INFO "scsi scan: 2nd INQUIRY"
	*/
	USBMS_DBG_SCAN(iprintf("[USBMS SCAN] scsi_probe_lun() : 2nd INQUIRY"
				" %s with code 0x%x\n", sreq->sr_result ?
				"failed" : "successful", sreq->sr_result));
	if (sreq->sr_result) {
		/* if the longer inquiry has failed, flag the device
		 * as only accepting 36 byte inquiries and retry the
		 * 36 byte inquiry */
		/*
		printk(KERN_INFO "scsi scan: %d byte inquiry failed"
			       " with code %d.  Consider BLIST_INQUIRY_36 for"
			       " this device\n", possible_inq_resp_len,
			       sreq->sr_result);
		*/
		dprintf("[USBMS] scsi scan: %d byte inquiry failed"
			       " with code %d.  Consider BLIST_INQUIRY_36 for"
			       " this device\n", possible_inq_resp_len,
			       sreq->sr_result);
		*bflags = BLIST_INQUIRY_36;

		USBMS_DBG_SCAN(iprintf(
			"[USBMS SCAN] scsi_probe_lun() : goto repeat_inquiry .. \n"
			)) ;
		// --------------
		goto repeat_inquiry;
		// --------------
	   }

	/*
	 * The INQUIRY can change, this means the length can change.
	 */
	possible_inq_resp_len = (unsigned char) inq_result[4] + 5;
	if (BLIST_INQUIRY_58 & *bflags)
			possible_inq_resp_len = 58;
	else if (possible_inq_resp_len > 255)
			possible_inq_resp_len = 36;	/* sanity */
	USBMS_DBG_INIT(iprintf(
		"[USBMS INIT] scsi_probe_lun() : possible_inq_resp_len=%d .. \n",
		possible_inq_resp_len
		));
   }

 sdev->inquiry_len = possible_inq_resp_len;
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] scsi_probe_lun() : Set Value sdev->inquiry_len=%d .. \n",
	sdev->inquiry_len
	));

 /*
  * XXX Abort if the response length is less than 36? If less than
  * 32, the lookup of the device flags (above) could be invalid,
  * and it would be possible to take an incorrect action - we do
  * not want to hang because of a short INQUIRY. On the flip side,
  * if the device is spun down or becoming ready (and so it gives a
  * short INQUIRY), an abort here prevents any further use of the
  * device, including spin up.
  *
  * Related to the above issue:
  *
  * XXX Devices (disk or all?) should be sent a TEST UNIT READY,
  * and if not ready, sent a START_STOP to start (maybe spin up) and
  * then send the INQUIRY again, since the INQUIRY can change after
  * a device is initialized.
  *
  * Ideally, start a device if explicitly asked to do so.  This
  * assumes that a device is spun up on power on, spun down on
  * request, and then spun up on request.
  */

 /*
  * The scanning code needs to know the scsi_level, even if no
  * device is attached at LUN 0 (SCSI_SCAN_TARGET_PRESENT) so
  * non-zero LUNs can be scanned.
  */
 sdev->scsi_level = inq_result[2] & 0x07;
 if (sdev->scsi_level >= 2 ||
    (sdev->scsi_level == 1 && (inq_result[3] & 0x0f) == 1))
		sdev->scsi_level++;
 /*
 #ifdef USBMS_DBG_BUFFER_ON   // I/O Buffer
   {
    unsigned char *buf_ptr ;

	buf_ptr =	(unsigned char *)(inq_result) ;
	BLOCK_PREEMPTION;
	USBMS_DBG_BUFFER(iprintf(
		"[USBMS BUFFER] scsi_probe_lun() : Dump inq_result "
		"Leading 32 Bytes : \n")) ;
	usb_DumpMem((char *) &(buf_ptr[0]), 32 ,USB_DUMP_DEFAULT|USB_DUMP_ASCII);
	UNBLOCK_PREEMPTION;
   }
 #endif // USBMS_DBG_BUFFER_ON
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_lun() End. \n"));
 return;
}

/**
 * scsi_add_lun - allocate and fully initialze a Scsi_Device
 * @sdevscan:	holds information to be stored in the new Scsi_Device
 * @sdevnew:	store the address of the newly allocated Scsi_Device
 * @inq_result:	holds the result of a previous INQUIRY to the LUN
 * @bflags:	black/white list flag
 *
 * Description:
 *     Allocate and initialize a Scsi_Device matching sdevscan. Optionally
 *     set fields based on values in *@bflags. If @sdevnew is not
 *     NULL, store the address of the new Scsi_Device in *@sdevnew (needed
 *     when scanning a particular LUN).
 *
 * Return:
 *     SCSI_SCAN_NO_RESPONSE: could not allocate or setup a Scsi_Device
 *     SCSI_SCAN_LUN_PRESENT: a new Scsi_Device was allocated and initialized
 **/
static int scsi_add_lun(struct scsi_device *sdev, char *inq_result, int *bflags)
{
 struct scsi_device *sdev_sibling;
 struct scsi_target *starget;
 unsigned long flags;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_lun() Begin .. \n")) ;
 /*
  * XXX do not save the inquiry, since it can change underneath us,
  * save just vendor/model/rev.
  *
  * Rather than save it and have an ioctl that retrieves the saved
  * value, have an ioctl that executes the same INQUIRY code used
  * in scsi_probe_lun, let user level programs doing INQUIRY
  * scanning run at their own risk, or supply a user level program
  * that can correctly scan.
  */

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_add_lun() : sdev->inquiry = UsbAllocMemC() .. \n"
		));

 /*
 sdev->inquiry = kmalloc(sdev->inquiry_len, GFP_ATOMIC);
 */
 sdev->inquiry = (unsigned char *) UsbAllocMemC(USB_MEM_C_S
					#ifdef USB_DBG_MEM_C_ON
						, sdev->inquiry_len
		    	    #endif
	           		) ;

 if (sdev->inquiry == NULL) {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] scsi_add_lun() : Error : SCSI_SCAN_NO_RESPONSE !! \n")) ;
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_lun() End with Fail !! \n")) ;
	return SCSI_SCAN_NO_RESPONSE;
   }

 #ifdef USBMS_DBG_BUFFER_ON   // I/O Buffer
   {
    unsigned char *buf_ptr ;

	buf_ptr = (unsigned char *)(inq_result) ;
	BLOCK_PREEMPTION;
	USBMS_DBG_BUFFER(iprintf(
		"[USBMS BUFFER] scsi_add_lun() : Dump inq_result "
		"Leading 32 Bytes : \n")) ;
	usb_DumpMem((char *) &(buf_ptr[0]), 32 ,USB_DUMP_DEFAULT);
	UNBLOCK_PREEMPTION;
   }
 #endif // USBMS_DBG_BUFFER_ON

 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] scsi_add_lun() : memcpy() sdev->inquiry_len=%d .. \n",
	sdev->inquiry_len
	));
 memcpy(sdev->inquiry, inq_result, sdev->inquiry_len);
 sdev->vendor = (char *) (sdev->inquiry + 8);
 sdev->model = (char *) (sdev->inquiry + 16);
 sdev->rev = (char *) (sdev->inquiry + 32);

 if (*bflags & BLIST_ISROM) {
		/*
		 * It would be better to modify sdev->type, and set
		 * sdev->removable, but then the print_inquiry() output
		 * would not show TYPE_ROM; if print_inquiry() is removed
		 * the issue goes away.
		 */
		inq_result[0] = TYPE_ROM;
		inq_result[1] |= 0x80;	/* removable */
	}

 switch (sdev->type = (inq_result[0] & 0x1f)) {
	case TYPE_TAPE:
	case TYPE_DISK:
	case TYPE_PRINTER:
	case TYPE_MOD:
	case TYPE_PROCESSOR:
	case TYPE_SCANNER:
	case TYPE_MEDIUM_CHANGER:
	case TYPE_ENCLOSURE:
	case TYPE_COMM:
		sdev->writeable = 1;
		break;
	case TYPE_WORM:
	case TYPE_ROM:
		sdev->writeable = 0;
		break;
	default:
		/*
		printk(KERN_INFO "scsi: unknown device type %d\n", sdev->type);
		*/
		dprintf("[USBMS] scsi: unknown device type %d\n", sdev->type);
	}

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_lun() : print_inquiry() .. \n")) ;
 print_inquiry(inq_result);

 /*
  * For a peripheral qualifier (PQ) value of 1 (001b), the SCSI
  * spec says: The device server is capable of supporting the
  * specified peripheral device type on this logical unit. However,
  * the physical device is not currently connected to this logical
  * unit.
  *
  * The above is vague, as it implies that we could treat 001 and
  * 011 the same. Stay compatible with previous code, and create a
  * Scsi_Device for a PQ of 1
  *
  * XXX Save the PQ field let the upper layers figure out if they
  * want to attach or not to this device, do not set online FALSE;
  * otherwise, offline devices still get an sd allocated, and they
  * use up an sd slot.
  */
 if (((inq_result[0] >> 5) & 7) == 1) {
	/*
	SCSI_LOG_SCAN_BUS(3, printk(KERN_INFO "scsi scan: peripheral"
	*/
	USBMS_DBG_SCAN(iprintf("[USBMS SCAN] scsi scan: peripheral"
				" qualifier of 1, device offlined\n"));
		sdev->online = FALSE;
	}

	sdev->removable = (0x80 & inq_result[1]) >> 7;
	sdev->lockable = sdev->removable;
	sdev->soft_reset = (inq_result[7] & 1) && ((inq_result[3] & 7) == 2);

	if (sdev->scsi_level >= SCSI_3 || (sdev->inquiry_len > 56 &&
		inq_result[56] & 0x04))
		sdev->ppr = 1;
	if (inq_result[7] & 0x60)
		sdev->wdtr = 1;
	if (inq_result[7] & 0x10)
		sdev->sdtr = 1;

 #ifdef CONFIG_USB_DEVICEFS
	sprintf(sdev->devfs_name, "scsi/host%d/bus%d/target%d/lun%d",
				sdev->host->host_no, sdev->channel,
				sdev->id, sdev->lun);

 #endif // CONFIG_USB_DEVICEFS
	/*
	 * End driverfs/devfs code.
	 */

	if ((sdev->scsi_level >= SCSI_2) && (inq_result[7] & 2) &&
	    !(*bflags & BLIST_NOTQ))
		sdev->tagged_supported = 1;
	/*
	 * Some devices (Texel CD ROM drives) have handshaking problems
	 * when used with the Seagate controllers. borken is initialized
	 * to 1, and then set it to 0 here.
	 */
	if ((*bflags & BLIST_BORKEN) == 0)
		sdev->borken = 0;

	/*
	 * Some devices may not want to have a start command automatically
	 * issued when a device is added.
	 */
	if (*bflags & BLIST_NOSTARTONADD)
		sdev->no_start_on_add = 1;

	/*
	 * If we need to allow I/O to only one of the luns attached to
	 * this target id at a time set single_lun, and allocate or modify
	 * sdev_target.
	 */
	if (*bflags & BLIST_SINGLELUN) {
		sdev->single_lun = 1;
 #ifndef NO_SPINLOCK
		spin_lock_irqsave(sdev->host->host_lock, flags);
#endif // ndef NO_SPINLOCK
		starget = NULL;
		/*
		 * Search for an existing target for this sdev.
		 */
		list_for_each_entry(sdev_sibling, &sdev->same_target_siblings,
				    same_target_siblings) {
			if (sdev_sibling->sdev_target != NULL) {
				starget = sdev_sibling->sdev_target;
				break;
			}
		}

		if (!starget) {
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_add_lun() : starget = UsbAllocMemC() .. \n"
		));

			/*
			starget = kmalloc(sizeof(*starget), GFP_ATOMIC);
			*/
			starget = (struct scsi_target *)
				UsbAllocMemC(USB_MEM_C_S
					#ifdef USB_DBG_MEM_C_ON
						, sizeof(struct scsi_target)
		    	    #endif
	           		) ;

			if (!starget) {
				/*
				printk(ALLOC_FAILURE_MSG, __FUNCTION__);
				*/
				dprintf(ALLOC_FAILURE_MSG);
				/*
				#ifndef NO_SPINLOCK
					spin_unlock_irqrestore(sdev->host->host_lock,
						       flags);
				#endif // ndef NO_SPINLOCK
				*/
				return SCSI_SCAN_NO_RESPONSE;
			}
			starget->starget_refcnt = 0;
			starget->starget_sdev_user = NULL;
		}
		starget->starget_refcnt++;
		sdev->sdev_target = starget;
 #ifndef NO_SPINLOCK
		spin_unlock_irqrestore(sdev->host->host_lock, flags);
#endif // ndef NO_SPINLOCK
	}

	/* if the device needs this changing, it may do so in the detect
	 * function */
	sdev->max_device_blocked = SCSI_DEFAULT_DEVICE_BLOCKED;

 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_add_lun() : Set ->use_10_for_rw=1 .. \n")) ;
 sdev->use_10_for_rw = 1;
 USBMS_DBG_INIT(iprintf("[USBMS INIT] scsi_add_lun() : Set ->use_10_for_ms=0 .. \n")) ;
 sdev->use_10_for_ms = 0;

 if(sdev->host->hostt->slave_configure)
		sdev->host->hostt->slave_configure(sdev);

	/*
	 * Ok, the device is now all set up, we can
	 * register it and tell the rest of the kernel
	 * about it.
	 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_lun() : scsi_device_register() .. \n")) ;
 scsi_device_register(sdev);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_lun() End. \n")) ;
 return SCSI_SCAN_LUN_PRESENT;
}

/**
 * scsi_probe_and_add_lun - probe a LUN, if a LUN is found add it
 * @sdevscan:	probe the LUN corresponding to this Scsi_Device
 * @sdevnew:	store the value of any new Scsi_Device allocated
 * @bflagsp:	store bflags here if not NULL
 *
 * Description:
 *     Call scsi_probe_lun, if a LUN with an attached device is found,
 *     allocate and set it up by calling scsi_add_lun.
 *
 * Return:
 *     SCSI_SCAN_NO_RESPONSE: could not allocate or setup a Scsi_Device
 *     SCSI_SCAN_TARGET_PRESENT: target responded, but no device is
 *         attached at the LUN
 *     SCSI_SCAN_LUN_PRESENT: a new Scsi_Device was allocated and initialized
 **/
static int scsi_probe_and_add_lun(struct Scsi_Host *host,
		uint channel, uint id, uint lun, int *bflagsp,
		struct scsi_device **sdevp, int rescan)
{
 struct scsi_device *sdev;
 struct scsi_request *sreq;
 unsigned char *result;
 int bflags, res = SCSI_SCAN_NO_RESPONSE;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_and_add_lun() Begin .. \n"));
 /*
  * The rescan flag is used as an optimization, the first scan of a
  * host adapter calls into here with rescan == 0.
  */
 if(rescan) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_and_add_lun() : scsi_find_device() .. \n"));
	sdev = scsi_find_device(host, channel, id, lun);
	if (sdev) {
		/*
		SCSI_LOG_SCAN_BUS(3, printk(KERN_INFO
		*/
		USBMS_DBG_SCAN(iprintf("[USBMS SCAN] "
				"scsi scan: device exists on <%d:%d:%d:%d>\n",
				host->host_no, channel, id, lun));
		if (sdevp)
				*sdevp = sdev;
		if (bflagsp)
				*bflagsp = scsi_get_device_flags(sdev->vendor,
								 sdev->model);
		return SCSI_SCAN_LUN_PRESENT;
	   }
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_and_add_lun() : scsi_alloc_sdev() .. \n"));
 sdev = scsi_alloc_sdev(host, channel, id, lun);
 USBMS_DBG_MEM(iprintf(
	"[USBMS MEM] scsi_probe_and_add_lun() : sdev=%08XH \n",
	sdev
	));
 if (!sdev)
		goto out;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_and_add_lun() : scsi_allocate_request() .. \n"));
 /*
 sreq = scsi_allocate_request(sdev, GFP_ATOMIC);
 */
 sreq = scsi_allocate_request(sdev);
 if (!sreq)
		goto out_free_sdev;

 // B. S. Liaw : request_buffer need uncached
 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_probe_and_add_lun() : result = UsbAllocMemU() .. \n"
		));
 /*
 result = kmalloc(256, GFP_ATOMIC |
			(host->unchecked_isa_dma) ? __GFP_DMA : 0);
 */
 result = (unsigned char *)	UsbAllocMemU(USB_MEM_U_M
				#ifdef USB_DBG_MEM_U_ON
					, 256
		        #endif
           		) ;

 if (!result)
		goto out_free_sreq;

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] scsi_probe_and_add_lun() : scsi_probe_lun() .. \n"
	));
 scsi_probe_lun(sreq, result, &bflags);
 /*
 #ifdef USBMS_DBG_BUFFER_ON   // I/O Buffer
   {
    unsigned char *buf_ptr ;

	buf_ptr = (unsigned char *)(result) ;
	BLOCK_PREEMPTION;
	USBMS_DBG_BUFFER(iprintf(
		"[USBMS BUFFER] scsi_probe_and_add_lun(1) : Dump result "
		"Leading 32 Bytes : \n")) ;
	usb_DumpMem((char *) &(buf_ptr[0]), 32 ,USB_DUMP_DEFAULT);
	UNBLOCK_PREEMPTION;
   }
 #endif // USBMS_DBG_BUFFER_ON
 */
 if(sreq->sr_result)
   {
	USBMS_DBG_INFO(iprintf("[USBMS FLOW] scsi_probe_and_add_lun() : scsi_probe_lun() Fail !! \n"));
	goto out_free_result;
   }

 /*
  * result contains valid SCSI INQUIRY data.
  */
 if ((result[0] >> 5) == 3) {
	/*
	 * For a Peripheral qualifier 3 (011b), the SCSI
	 * spec says: The device server is not capable of
	 * supporting a physical device on this logical
	 * unit.
	 *
	 * For disks, this implies that there is no
	 * logical disk configured at sdev->lun, but there
	 * is a target id responding.
	 */
	/*
	SCSI_LOG_SCAN_BUS(3, printk(KERN_INFO
	*/
	USBMS_DBG_SCAN(iprintf("[USBMS SCAN] " 
					"scsi scan: peripheral qualifier of 3,"
					" no device added\n"));
	res = SCSI_SCAN_TARGET_PRESENT;
	goto out_free_result;
   }
 /*
 #ifdef USBMS_DBG_BUFFER_ON   // I/O Buffer
   {
    unsigned char *buf_ptr ;

	buf_ptr = (unsigned char *)(result) ;
	BLOCK_PREEMPTION;
	USBMS_DBG_BUFFER(iprintf(
		"[USBMS BUFFER] scsi_probe_and_add_lun(2) : Dump result "
		"Leading 32 Bytes : \n")) ;
	usb_DumpMem((char *) &(buf_ptr[0]), 32 ,USB_DUMP_DEFAULT);
	UNBLOCK_PREEMPTION;
   }
 #endif // USBMS_DBG_BUFFER_ON
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_and_add_lun() : scsi_add_lun() .. \n"));
 res = scsi_add_lun(sdev, result, &bflags);
 if(res == SCSI_SCAN_LUN_PRESENT) {
	if(bflags & BLIST_KEY) {
		sdev->lockable = 0;
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_and_add_lun() : scsi_unlock_floptical() .. \n"));
		scsi_unlock_floptical(sreq, result);
	   }
	if (bflagsp)
		*bflagsp = bflags;
   }

 out_free_result:

 USBMS_DBG_MEM(iprintf("[USBMS MEM] scsi_probe_and_add_lun() : UsbFreeMemU(result) .. \n"));
 /*
 kfree(result);
 */
 UsbFreeMemU(USB_MEM_U_M,(unsigned char *) result
				#if defined(USB_DBG_MEM_U_ON)
					, 256
				#endif
				);

 out_free_sreq:
 scsi_release_request(sreq);
 out_free_sdev:
 if(res == SCSI_SCAN_LUN_PRESENT) {
	if (sdevp)
			*sdevp = sdev;
   } else {
	scsi_free_sdev(sdev);
   }
 out:

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_probe_and_add_lun() End. \n"));
 return res;
}

/**
 * scsi_sequential_lun_scan - sequentially scan a SCSI target
 * @sdevscan:	scan the host, channel, and id of this Scsi_Device
 * @bflags:	black/white list flag for LUN 0
 * @lun0_res:	result of scanning LUN 0
 *
 * Description:
 *     Generally, scan from LUN 1 (LUN 0 is assumed to already have been
 *     scanned) to some maximum lun until a LUN is found with no device
 *     attached. Use the bflags to figure out any oddities.
 *
 *     Modifies sdevscan->lun.
 **/
static void scsi_sequential_lun_scan(struct Scsi_Host *shost, uint channel,
		uint id, int bflags, int lun0_res, int scsi_level, int rescan)
{
 unsigned int sparse_lun, lun, max_dev_lun;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sequential_lun_scan() Begin .. \n"));
 /*
 SCSI_LOG_SCAN_BUS(3, printk(KERN_INFO "scsi scan: Sequential scan of"
 */
 USBMS_DBG_SCAN(iprintf("[USBMS SCAN] Sequential scan of"
			" host %d channel %d id %d\n", shost->host_no,
			channel, id));

 USBMS_DBG_SCAN(iprintf("[USBMS SCAN] scsi_sequential_lun_scan() : "
	" max_scsi_luns=%d \n", max_scsi_luns
	));
 USBMS_DBG_SCAN(iprintf("[USBMS SCAN] scsi_sequential_lun_scan() : "
	" shost->max_lun=%d \n", shost->max_lun
	));
 max_dev_lun = min(max_scsi_luns, shost->max_lun);
 /*
  * If this device is known to support sparse multiple units,
  * override the other settings, and scan all of them. Normally,
  * SCSI-3 devices should be scanned via the REPORT LUNS.
  */
 if (bflags & BLIST_SPARSELUN) {
		max_dev_lun = shost->max_lun;
		sparse_lun = 1;
	} else
		sparse_lun = 0;

 /*
  * If not sparse lun and no device attached at LUN 0 do not scan
  * any further.
  */
 if (!sparse_lun && (lun0_res != SCSI_SCAN_LUN_PRESENT))
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sequential_lun_scan(1) End with Failed !! \n"));
	return;
   }

 /*
  * If less than SCSI_1_CSS, and no special lun scaning, stop
  * scanning; this matches 2.4 behaviour, but could just be a bug
  * (to continue scanning a SCSI_1_CSS device).
  *
  * This test is broken.  We might not have any device on lun0 for
  * a sparselun device, and if that's the case then how would we
  * know the real scsi_level, eh?  It might make sense to just not
  * scan any SCSI_1 device for non-0 luns, but that check would best
  * go into scsi_alloc_sdev() and just have it return null when asked
  * to alloc an sdev for lun > 0 on an already found SCSI_1 device.
  *
 if ((sdevscan->scsi_level < SCSI_1_CCS) &&
	    ((bflags & (BLIST_FORCELUN | BLIST_SPARSELUN | BLIST_MAX5LUN))
	     == 0))
		return;
  */
 /*
  * If this device is known to support multiple units, override
  * the other settings, and scan all of them.
  */
 if (bflags & BLIST_FORCELUN)
		max_dev_lun = shost->max_lun;
 /*
  * REGAL CDC-4X: avoid hang after LUN 4
  */
 if (bflags & BLIST_MAX5LUN)
		max_dev_lun = min(5U, max_dev_lun);
 /*
  * Do not scan SCSI-2 or lower device past LUN 7, unless
  * BLIST_LARGELUN.
  */
 if (scsi_level < SCSI_3 && !(bflags & BLIST_LARGELUN))
		max_dev_lun = min(8U, max_dev_lun);

 /*
  * We have already scanned LUN 0, so start at LUN 1. Keep scanning
  * until we reach the max, or no LUN is found and we are not
  * sparse_lun.
  */
 USBMS_DBG_SCAN(iprintf("[USBMS SCAN] scsi_sequential_lun_scan() : "
	" max_dev_lun=%d \n", max_dev_lun
	));

 #ifndef USBMS_ONLY_ONE_DISK
	#ifdef USBMS_BS_ASSERT
		if ( !(    max_dev_lun
			    >= 2
		   )  ) {
			iprintf("[ASSERT][USBMS] scsi_sequential_lun_scan() : " ) ;
		    iprintf("Error : Invalid max_dev_lun !!\n" ) ;
		    iprintf("[ASSERT][USBMS] max_dev_lun=%d \n", max_dev_lun ) ;
	    	iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	#endif // USBMS_BS_ASSERT
 #endif // ndef USBMS_ONLY_ONE_DISK

 for (lun = 1; lun < max_dev_lun; ++lun)
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sequential_lun_scan(2) : scsi_probe_and_add_lun() .. \n"));
	if ((scsi_probe_and_add_lun(shost, channel, id, lun,
		      NULL, NULL, rescan) != SCSI_SCAN_LUN_PRESENT) &&
		    !sparse_lun) {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sequential_lun_scan(3) End with Failed !! \n"));
		return;
	   }
   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_sequential_lun_scan() End. \n"));
}

#ifdef CONFIG_SCSI_REPORT_LUNS
/**
 * scsilun_to_int: convert a scsi_lun to an int
 * @scsilun:	struct scsi_lun to be converted.
 *
 * Description:
 *     Convert @scsilun from a struct scsi_lun to a four byte host byte-ordered
 *     integer, and return the result. The caller must check for
 *     truncation before using this function.
 *
 * Notes:
 *     The struct scsi_lun is assumed to be four levels, with each level
 *     effectively containing a SCSI byte-ordered (big endian) short; the
 *     addressing bits of each level are ignored (the highest two bits).
 *     For a description of the LUN format, post SCSI-3 see the SCSI
 *     Architecture Model, for SCSI-3 see the SCSI Controller Commands.
 *
 *     Given a struct scsi_lun of: 0a 04 0b 03 00 00 00 00, this function returns
 *     the integer: 0x0b030a04
 **/
static int scsilun_to_int(struct scsi_lun *scsilun)
{
	int i;
	unsigned int lun;

	lun = 0;
	for (i = 0; i < sizeof(lun); i += 2)
		lun = lun | (((scsilun->scsi_lun[i] << 8) |
			      scsilun->scsi_lun[i + 1]) << (i * 8));
	return lun;
}

/**
 * scsi_report_lun_scan - Scan using SCSI REPORT LUN results
 * @sdevscan:	scan the host, channel, and id of this Scsi_Device
 *
 * Description:
 *     If @sdevscan is for a SCSI-3 or up device, send a REPORT LUN
 *     command, and scan the resulting list of LUNs by calling
 *     scsi_probe_and_add_lun.
 *
 *     Modifies sdevscan->lun.
 *
 * Return:
 *     0: scan completed (or no memory, so further scanning is futile)
 *     1: no report lun scan, or not configured
 **/
static int scsi_report_lun_scan(struct scsi_device *sdev, int bflags,
				int rescan)
{
 char devname[64];
 unsigned char scsi_cmd[MAX_COMMAND_SIZE];
 unsigned int length;
 unsigned int lun;
 unsigned int num_luns;
 unsigned int retries;
 struct scsi_lun *lunp, *lun_data;
 struct scsi_request *sreq;
 char *data;

 /*
  * Only support SCSI-3 and up devices.
  */
 if (sdev->scsi_level < SCSI_3)
		return 1;
 if (bflags & BLIST_NOLUN)
		return 0;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_report_lun_scan() : scsi_allocate_request() .. \n"));
 /*
 sreq = scsi_allocate_request(sdev, GFP_ATOMIC);
 */
 sreq = scsi_allocate_request(sdev);
	if (!sreq)
		goto out;



 sprintf(devname, "host %d channel %d id %d",
		sdev->host->host_no, sdev->channel, sdev->id);

 /*
  * Allocate enough to hold the header (the same size as one scsi_lun)
  * plus the max number of luns we are requesting.
  *
  * Reallocating and trying again (with the exact amount we need)
  * would be nice, but then we need to somehow limit the size
  * allocated based on the available memory and the limits of
  * kmalloc - we don't want a kmalloc() failure of a huge value to
  * prevent us from finding any LUNs on this target.
  */
 length = (max_scsi_report_luns + 1) * sizeof(struct scsi_lun);

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_report_lun_scan() : lun_data = UsbAllocMemU() .. \n"
		));

 // B. S. Liaw : request_buffer need uncached
 /*
 lun_data = kmalloc(length, GFP_ATOMIC |
			   (sdev->host->unchecked_isa_dma ? __GFP_DMA : 0));
 */
 lun_data = (struct scsi_lun *) UsbAllocMemU(USB_MEM_U_M
				#ifdef USB_DBG_MEM_C_ON
					, length            // if 128 : size : 1032
		        #endif
           		) ;
 if(!lun_data)
   {
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] scsi_report_lun_scan() : Error : Out of Memory for lun_data !! \n"
		));
	dprintf("[USBMS] Error : Out of Memory !! \n");
	goto out_release_request;
   }

 scsi_cmd[0] = REPORT_LUNS;

 /*
  * bytes 1 - 5: reserved, set to zero.
  */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_report_lun_scan() : memset() .. \n" )) ;
 memset(&scsi_cmd[1], 0, 5);

 /*
  * bytes 6 - 9: length of the command.
  */
 scsi_cmd[6] = (unsigned char) (length >> 24) & 0xff;
 scsi_cmd[7] = (unsigned char) (length >> 16) & 0xff;
 scsi_cmd[8] = (unsigned char) (length >> 8) & 0xff;
 scsi_cmd[9] = (unsigned char) length & 0xff;

 scsi_cmd[10] = 0;	/* reserved */
 scsi_cmd[11] = 0;	/* control */
 sreq->sr_cmd_len = 0;
 sreq->sr_data_direction = DMA_FROM_DEVICE;

 /*
  * We can get a UNIT ATTENTION, for example a power on/reset, so
  * retry a few times (like sd.c does for TEST UNIT READY).
  * Experience shows some combinations of adapter/devices get at
  * least two power on/resets.
  *
  * Illegal requests (for devices that do not support REPORT LUNS)
  * should come through as a check condition, and will not generate
  * a retry.
  */
 for (retries = 0; retries < 3; retries++) {
	/*
	SCSI_LOG_SCAN_BUS(3, printk (KERN_INFO "scsi scan: Sending"
	*/
	USBMS_DBG_SCAN(iprintf("[USBMS SCAN] scsi_report_lun_scan() : Sending"
				" REPORT LUNS to %s (try %d)\n", devname,
				retries));
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_report_lun_scan() : scsi_wait_req() .. \n"));
	scsi_wait_req(sreq, scsi_cmd, lun_data, length,
				/*
				SCSI_TIMEOUT + 4*HZ, 3);
				*/
				SCSI_TIMEOUT + 4 , 3);
	/*
	SCSI_LOG_SCAN_BUS(3, printk (KERN_INFO "scsi scan: REPORT LUNS"
	*/
	USBMS_DBG_SCAN(iprintf("[USBMS SCAN] scsi scan: REPORT LUNS"
				" %s (try %d) result 0x%x\n", sreq->sr_result
				?  "failed" : "successful", retries,
				sreq->sr_result));
	
		if (sreq->sr_result == 0 ||
		    sreq->sr_sense_buffer[2] != UNIT_ATTENTION)
			break;
	}

	if (sreq->sr_result) {
		/*
		 * The device probably does not support a REPORT LUN command
		 */
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_report_lun_scan() : UsbFreeMemU(lun_data) .. \n"
		));
	/*
	kfree(lun_data);
	*/
	UsbFreeMemU(USB_MEM_U_M,(unsigned char *) lun_data
				#if defined(USB_DBG_MEM_U_ON)
					,sizeof(*lun_data)   // if 128 : size : 1032
				#endif
				) ;
		scsi_release_request(sreq);
		return 1;
	}
	scsi_release_request(sreq);

	/*
	 * Get the length from the first four bytes of lun_data.
	 */
	data = (char *) lun_data->scsi_lun;
	length = ((data[0] << 24) | (data[1] << 16) |
		  (data[2] << 8) | (data[3] << 0));

	num_luns = (length / sizeof(struct scsi_lun));
	if (num_luns > max_scsi_report_luns) {
		/*
		printk(KERN_WARNING "scsi: On %s only %d (max_scsi_report_luns)"
		       " of %d luns reported, try increasing"
		       " max_scsi_report_luns.\n", devname,
		       max_scsi_report_luns, num_luns);
		*/
		dprintf("scsi: On %s only %d (max_scsi_report_luns)"
		       " of %d luns reported, try increasing"
		       " max_scsi_report_luns.\n", devname,
		       max_scsi_report_luns, num_luns);
		num_luns = max_scsi_report_luns;
	}
 /*
 SCSI_LOG_SCAN_BUS(3, printk (KERN_INFO "scsi scan: REPORT LUN scan of"
 */
 USBMS_DBG_SCAN(iprintf("[USBMS SCAN] scsi scan: REPORT LUN scan of"
	" host %d channel %d id %d\n", sdev->host->host_no,
	sdev->channel, sdev->id));

	/*
	 * Scan the luns in lun_data. The entry at offset 0 is really
	 * the header, so start at 1 and go up to and including num_luns.
	 */
	for (lunp = &lun_data[1]; lunp <= &lun_data[num_luns]; lunp++) {
		lun = scsilun_to_int(lunp);

		/*
		 * Check if the unused part of lunp is non-zero, and so
		 * does not fit in lun.
		 */
		if (memcmp(&lunp->scsi_lun[sizeof(lun)], "\0\0\0\0", 4)) {
			int i;

			/*
			 * Output an error displaying the LUN in byte order,
			 * this differs from what linux would print for the
			 * integer LUN value.
			 */
			/*
			printk(KERN_WARNING "scsi: %s lun 0x", devname);
			*/
			dprintf("scsi: %s lun 0x", devname);
			data = (char *)lunp->scsi_lun;
			for (i = 0; i < sizeof(struct scsi_lun); i++)
				printk("%02x", data[i]);
			printk(" has a LUN larger than currently supported.\n");
		} else if (lun == 0) {
			/*
			 * LUN 0 has already been scanned.
			 */
		} else if (lun > sdev->host->max_lun) {
			/*
			printk(KERN_WARNING "scsi: %s lun%d has a LUN larger"
			       " than allowed by the host adapter\n",
			       devname, lun);
			*/
			dprintf("scsi: %s lun%d has a LUN larger"
			       " than allowed by the host adapter\n",
			       devname, lun);
		} else {
			int res;

			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_report_lun_scan() : scsi_probe_and_add_lun() .. \n"));
			res = scsi_probe_and_add_lun(sdev->host, sdev->channel,
				sdev->id, lun, NULL, NULL, rescan);
			if (res == SCSI_SCAN_NO_RESPONSE) {
				/*
				 * Got some results, but now none, abort.
				 */
				/*
				printk(KERN_ERR "scsi: Unexpected response"
				       " from %s lun %d while scanning, scan"
				       " aborted\n", devname, lun);
				*/
				dprintf("[USBMS] Error :scsi: Unexpected response"
				       " from %s lun %d while scanning, scan"
				       " aborted\n", devname, lun);
				break;
			}
		}
	}

 USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] scsi_report_lun_scan() : UsbFreeMemU(lun_data) .. \n"
		));
 /*
 kfree(lun_data);
 */
 UsbFreeMemU(USB_MEM_U_M,(unsigned char *) lun_data
				#if defined(USB_DBG_MEM_U_ON)
					,sizeof(*lun_data)      // if 128 : size : 1032
				#endif
				) ;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_report_lun_scan() End. \n" )) ;
 return 0;

 out_release_request:
	scsi_release_request(sreq);
 out:
 /*
  * We are out of memory, don't try scanning any further.
  */
 /*
 printk(ALLOC_FAILURE_MSG, __FUNCTION__);
 */
 dprintf(ALLOC_FAILURE_MSG);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_report_lun_scan() End with Failed !! \n" )) ;
 return 0;
}
#else
# define scsi_report_lun_scan(sdev, blags, rescan)	(1)
#endif	/* CONFIG_SCSI_REPORT_LUNS */

struct scsi_device *scsi_add_device(struct Scsi_Host *shost,
				    uint channel, uint id, uint lun)
{
 struct scsi_device *sdev;
 int res;


 #ifndef NO_SEMPHORE
	down(&shost->scan_mutex);
 #endif // ndef NO_SEMPHORE
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_add_device() : scsi_probe_and_add_lun() .. \n"));
 res = scsi_probe_and_add_lun(shost, channel, id, lun, NULL, &sdev, 1);
 if (res != SCSI_SCAN_LUN_PRESENT)
	   {
		#ifndef _ETCPIP
			sdev = ERR_PTR(-ENODEV);
		 #else
			sdev = (struct scsi_device *) 0x00L ;
		#endif
	   }
 #ifndef NO_SEMPHORE
	up(&shost->scan_mutex);
 #endif // ndef NO_SEMPHORE

	return sdev;
}

void scsi_rescan_device(struct device *dev)
{
	struct scsi_driver *drv = to_scsi_driver(dev->driver);

	if (try_module_get(drv->owner)) {
		if (drv->rescan)
			drv->rescan(dev);
		module_put(drv->owner);
	}
}

/**
 * scsi_scan_target - scan a target id, possibly including all LUNs on the
 *     target.
 * @sdevsca:	Scsi_Device handle for scanning
 * @shost:	host to scan
 * @channel:	channel to scan
 * @id:		target id to scan
 *
 * Description:
 *     Scan the target id on @shost, @channel, and @id. Scan at least LUN
 *     0, and possibly all LUNs on the target id.
 *
 *     Use the pre-allocated @sdevscan as a handle for the scanning. This
 *     function sets sdevscan->host, sdevscan->id and sdevscan->lun; the
 *     scanning functions modify sdevscan->lun.
 *
 *     First try a REPORT LUN scan, if that does not scan the target, do a
 *     sequential scan of LUNs on the target id.
 **/
static void scsi_scan_target(struct Scsi_Host *shost, unsigned int channel,
			     unsigned int id, unsigned int lun, int rescan)
{
 int bflags = 0;
 int res;
 struct scsi_device *sdev;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_target() Begin .. \n"));

 if (shost->this_id == id)
   {
	/*
	 * Don't scan the host adapter
	 */
	USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_scan_target() : Don't scan the host adapter \n"));
	return;
   }
 if (lun != SCAN_WILD_CARD) {
		/*
		 * Scan for a specific host/chan/id/lun.
		 */
	USBMS_DBG_INFO(iprintf("[USBMS INFO] scsi_scan_target() : Scan for a specific host/chan/id/lun \n"));
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_target() : scsi_probe_and_add_lun() .. \n"));
	scsi_probe_and_add_lun(shost, channel, id, lun, NULL, NULL,
				       rescan);
	return;
   }

 /*
  * Scan LUN 0, if there is some response, scan further. Ideally, we
  * would not configure LUN 0 until all LUNs are scanned.
  */
 USBMS_DBG_SCAN(iprintf("[USBMS SCAN] scsi_scan_target() : Scan LUN 0 \n"));
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_target() : scsi_probe_and_add_lun() .. \n"));
 res = scsi_probe_and_add_lun(shost, channel, id, 0, &bflags, &sdev,
				     rescan);

 if (res == SCSI_SCAN_LUN_PRESENT) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_target() : scsi_report_lun_scan() .. \n"));
	if (scsi_report_lun_scan(sdev, bflags, rescan) != 0)
		/*
		 * The REPORT LUN did not scan the target,
		 * do a sequential scan.
		 */
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_target() : scsi_sequential_lun_scan() .. \n"));
		scsi_sequential_lun_scan(shost, channel, id, bflags,
				       	res, sdev->scsi_level, rescan);
	   } else if (res == SCSI_SCAN_TARGET_PRESENT) {
		/*
		 * There's a target here, but lun 0 is offline so we
		 * can't use the report_lun scan.  Fall back to a
		 * sequential lun scan with a bflags of SPARSELUN and
		 * a default scsi level of SCSI_2
		 */
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_target() : scsi_sequential_lun_scan() .. \n"));
		scsi_sequential_lun_scan(shost, channel, id, BLIST_SPARSELUN,
				SCSI_SCAN_TARGET_PRESENT, SCSI_2, rescan);
	   }

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_target() End. \n"));
}

static void scsi_scan_channel(struct Scsi_Host *shost, unsigned int channel,
			      unsigned int id, unsigned int lun, int rescan)
{
 uint order_id;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_channel() Begin .. \n")) ;
 if (id == SCAN_WILD_CARD)
	for(id = 0; id < shost->max_id; ++id) {
		/*
		 * XXX adapter drivers when possible (FCP, iSCSI)
		 * could modify max_id to match the current max,
		 * not the absolute max.
		 *
		 * XXX add a shost id iterator, so for example,
		 * the FC ID can be the same as a target id
		 * without a huge overhead of sparse id's.
		 */
		if (shost->reverse_ordering)
				/*
				 * Scan from high to low id.
				 */
				order_id = shost->max_id - id - 1;
			else
				order_id = id;

		USBMS_DBG_SCAN(iprintf("[USBMS SCAN] scsi_scan_channel() : id=%d .. \n",id)) ;
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_channel(1) : scsi_scan_target() .. \n")) ;
		scsi_scan_target(shost, channel, order_id, lun, rescan);
	   }
  else
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_channel(2) : scsi_scan_target() .. \n")) ;
	scsi_scan_target(shost, channel, id, lun, rescan);
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_channel() End. \n")) ;
}

int scsi_scan_host_selected(struct Scsi_Host *shost, unsigned int channel,
			    unsigned int id, unsigned int lun, int rescan)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_host_selected() Begin .. \n")) ;
 /*
 SCSI_LOG_SCAN_BUS(3, printk (KERN_INFO "%s: <%u:%u:%u:%u>\n",
 */
 USBMS_DBG_SCAN(iprintf(
	"[USBMS SCAN] scsi_scan_host_selected() : <%u:%u:%u:%u>\n",
	shost->host_no, channel, id, lun));

 if(((channel != SCAN_WILD_CARD) && (channel > shost->max_channel)) ||
	((id != SCAN_WILD_CARD) && (id > shost->max_id)) ||
	((lun != SCAN_WILD_CARD) && (lun > shost->max_lun)))
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_host_selected() End with Fail !! \n")) ;
	return -EINVAL;
   }

 #ifndef NO_SEMPHORE
	down(&shost->scan_mutex);
 #endif // ndef NO_SEMPHORE

 if (channel == SCAN_WILD_CARD) 
	for (channel = 0; channel <= shost->max_channel; channel++)
	   {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_host_selected() : scsi_scan_channel() .. \n")) ;
		scsi_scan_channel(shost, channel, id, lun, rescan);
	   }
  else
   {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_host_selected() : scsi_scan_channel() .. \n")) ;
	scsi_scan_channel(shost, channel, id, lun, rescan);
   }

 #ifndef NO_SEMPHORE
	up(&shost->scan_mutex);
 #endif // ndef NO_SEMPHORE

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_host_selected() End. \n")) ;
 return 0;
}

/**
 * scsi_scan_host - scan the given adapter
 * @shost:	adapter to scan
 **/
void scsi_scan_host(struct Scsi_Host *shost)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_scan_host() : scsi_scan_host_selected() .. \n")) ;
 scsi_scan_host_selected(shost, SCAN_WILD_CARD, SCAN_WILD_CARD,
				SCAN_WILD_CARD, 0);
}

void scsi_forget_host(struct Scsi_Host *shost)
{
 struct list_head *le, *lh;
 struct scsi_device *sdev;


 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_forget_host() Begin .. \n"));
 #ifndef _ETCPIP
	list_for_each_safe(le, lh, &shost->my_devices) {
  #else
	list_for_each_safe(le, lh, shost->my_devices) {
 #endif
		sdev = list_entry(le, struct scsi_device, siblings);
		

		USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_forget_host() : sdev=%08XH \n",sdev));

		#ifdef _ETCPIP
			USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_forget_host() : &sdev->sdev_gendev=%08XH \n",
				&sdev->sdev_gendev));
			USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_forget_host() : device_del() .. \n"));
			device_del(&sdev->sdev_gendev);
		#endif // _ETCPIP

		USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_forget_host() : scsi_remove_device() .. \n"));
		scsi_remove_device(sdev);
		#ifdef _ETCPIP
			USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_forget_host() : scsi_free_sdev() .. \n"));
			scsi_free_sdev(sdev);
		#endif // _ETCPIP
	   }
}

/*
 * Function:    scsi_get_host_dev()
 *
 * Purpose:     Create a Scsi_Device that points to the host adapter itself.
 *
 * Arguments:   SHpnt   - Host that needs a Scsi_Device
 *
 * Lock status: None assumed.
 *
 * Returns:     The Scsi_Device or NULL
 *
 * Notes:
 *	Attach a single Scsi_Device to the Scsi_Host - this should
 *	be made to look like a "pseudo-device" that points to the
 *	HA itself.
 *
 *	Note - this device is not accessible from any high-level
 *	drivers (including generics), which is probably not
 *	optimal.  We can add hooks later to attach 
 */
struct scsi_device *scsi_get_host_dev(struct Scsi_Host *shost)
{
	struct scsi_device *sdev;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] scsi_get_host_dev() : scsi_alloc_sdev() .. \n"));
	sdev = scsi_alloc_sdev(shost, 0, shost->this_id, 0);
	if (sdev) {
		sdev->borken = 0;
	}
	return sdev;
}

/*
 * Function:    scsi_free_host_dev()
 *
 * Purpose:     Free a scsi_device that points to the host adapter itself.
 *
 * Arguments:   SHpnt   - Host that needs a Scsi_Device
 *
 * Lock status: None assumed.
 *
 * Returns:     Nothing
 *
 * Notes:
 */
void scsi_free_host_dev(struct scsi_device *sdev)
{
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] scsi_free_host_dev() Begin .. \n"));
 /*
 BUG_ON(sdev->id != sdev->host->this_id);
 */
 if(sdev->id != sdev->host->this_id) {
	dprintf("[USBMS] BUG : Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__);
   }
 scsi_free_sdev(sdev);
}

// B. S. Liaw TEST for compiler warning !!

